# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
import dataclasses
import enum

from gt4py._core import definitions as core_defs
from gt4py.eve.extended_typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    Mapping,
    MaybeNestedInTuple,
    Optional,
    Self,
    TypeAlias,
    TypeIs,
    TypeVar,
    cast,
    final,
)
from gt4py.next import common, errors, named_collections, utils
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation


def _make_dict_expr(exprs: dict[str, str]) -> str:
    items = str.join(",", (f"'{k}': {v}" for k, v in exprs.items()))
    return f"{{{items}}}"


class ArgStaticDescriptor(abc.ABC):
    """
    Abstract class to represent, extract, validate compile time information of an argument.

    The information that is available at compile time is extracted from the runtime argument
    (or provided when pre-compiling) is described by a set of (python) expressions returned by the
    `attribute_extractor` class-method. These expressions are evaluated in the context of the
    arguments. We chose expressions here instead of a method taking the actual value such that we
    can code generate a single expression for all argument descriptors only retrieving the necessary
    values without actually constructing the descriptors. That way the cache key computation to the
    compiled is fast.
    """

    def validate(self, name: str, type_: ts.TypeSpec) -> None:  # noqa: B027  # method is not abstract, but just empty when not implemented
        """
        Validate argument descriptor in the context of an actual program.

        This function is called when the type of the argument is available. The name is merely
        given to give good error messages.
        """
        pass

    @classmethod
    @final
    def from_value(cls, value: Any) -> ArgStaticDescriptor:
        attr_exprs = cls.attribute_extractor_exprs("self")
        return cls(**eval(f"""lambda self: {_make_dict_expr(attr_exprs)}""")(value))

    @classmethod
    @abc.abstractmethod
    def attribute_extractor_exprs(cls, arg_expr: str) -> dict[str, str]:
        """
        Return a mapping from the attributes of our descriptor to the expressions to retrieve them.

        E.g. if `arg_expr` would be `myarg` and the result of this function
        `{'value': 'my_arg.value'}` then the descriptor is constructed as
        `ArgumentDescriptor(value=my_arg.value)`. We use an expression here such that we can compute
        a cache key by just hashing `my_arg.value` instead of first constructing the descriptor.
        """
        ...


@dataclasses.dataclass(frozen=True)
class StaticArg(ArgStaticDescriptor, Generic[core_defs.ScalarT]):
    value: MaybeNestedInTuple[core_defs.ScalarT]

    def __post_init__(self) -> None:
        # transform enum value into the actual value
        object.__setattr__(
            self,
            "value",
            utils.tree_map(lambda val: val.value if isinstance(val, enum.Enum) else val)(
                self.value
            ),
        )

    def validate(self, name: str, type_: ts.TypeSpec) -> None:
        if not type_info.is_type_or_tuple_of_type(type_, ts.ScalarType):
            raise errors.DSLTypeError(
                message=f"Invalid static argument '{name}' with type '{type_}' (only scalars or (nested) tuples of scalars can be static).",
                location=None,
            )

        actual_type = type_translation.from_value(self.value)
        if actual_type != type_:
            raise errors.DSLTypeError(
                message=f"Invalid static argument '{name}', expected '{type_}', but static value '{self.value}' has type '{actual_type}'.",
                location=None,
            )

    @classmethod
    def attribute_extractor_exprs(cls, arg_expr: str) -> dict[str, str]:
        return {"value": arg_expr}


@dataclasses.dataclass(frozen=True)
class FieldDomainDescriptor(ArgStaticDescriptor):
    domain: common.Domain

    @classmethod
    def attribute_extractor_exprs(cls, arg_expr: str) -> dict[str, str]:
        return {"domain": f"({arg_expr}).domain"}


@dataclasses.dataclass(frozen=True)
class JITArgs:
    """Concrete (runtime) arguments to a GTX program in a format that can be passed into the toolchain."""

    args: tuple[Any, ...]
    kwargs: dict[str, Any]

    @classmethod
    def from_signature(cls, *args: Any, **kwargs: Any) -> Self:
        return cls(args=args, kwargs=kwargs)


ArgStaticDescriptorsContext: TypeAlias = dict[str, MaybeNestedInTuple[ArgStaticDescriptor | None]]
ArgStaticDescriptorsContextsByType: TypeAlias = Mapping[
    type[ArgStaticDescriptor], ArgStaticDescriptorsContext
]


@dataclasses.dataclass(frozen=True)
class CompileTimeArgs:
    """Compile-time standins for arguments to a GTX program to be used in ahead-of-time compilation."""

    args: tuple[ts.TypeSpec, ...]
    kwargs: dict[str, ts.TypeSpec]
    offset_provider: common.OffsetProvider  # TODO(havogt): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime information
    column_axis: Optional[common.Dimension]
    #: A mapping from an argument descriptor type to a context containing the actual descriptors.
    #: If an argument or element of an argument has no descriptor, the respective value is `None`.
    #: E.g., for a tuple argument `a` with type `ts.TupleTupe(types=[field_t, int32_t])` a possible
    #  context would be `{"a": (FieldDomainDescriptor(...), None)}`.
    argument_descriptor_contexts: ArgStaticDescriptorsContextsByType

    @property
    def offset_provider_type(self) -> common.OffsetProviderType:
        return common.offset_provider_to_type(self.offset_provider)

    @classmethod
    def from_concrete(cls, *args: Any, **kwargs: Any) -> Self:
        """Convert concrete GTX program arguments into their compile-time counterparts."""
        kwargs_copy = kwargs.copy()
        return cls(
            args=tuple(type_translation.from_value(arg) for arg in args),
            offset_provider=kwargs_copy.pop("offset_provider", {}),
            column_axis=kwargs_copy.pop("column_axis", None),
            kwargs={
                k: type_translation.from_value(v) for k, v in kwargs_copy.items() if v is not None
            },
            argument_descriptor_contexts={},
        )

    @classmethod
    def empty(cls) -> Self:
        return cls(tuple(), {}, {}, None, {})


# This is not really accurate, just an approximation
NeedsValueExtraction: TypeAlias = MaybeNestedInTuple[named_collections.CustomNamedCollection]


def needs_value_extraction(value: object) -> TypeIs[NeedsValueExtraction]:
    return isinstance(value, named_collections.CUSTOM_NAMED_COLLECTION_TYPES) or (
        isinstance(value, tuple) and any(needs_value_extraction(v) for v in value)
    )


T = TypeVar("T")


def extract(
    value: NeedsValueExtraction | MaybeNestedInTuple[common.PrimitiveValue],
    pass_through_values: bool = True,
) -> MaybeNestedInTuple[common.PrimitiveValue]:
    """
    Extract the values from a run-time argument into a digestible value form.

    This functions is useful to do the extraction from run-time values
    when the full type information is not available. For primitive values,
    return them as-is if `pass_through_values` is `True`, otherwise raise a `TypeError`.
    """
    if isinstance(value, common.PRIMITIVE_VALUE_TYPES):
        return cast(common.PrimitiveValue, value)
    if isinstance(value, named_collections.CUSTOM_NAMED_COLLECTION_TYPES):
        return named_collections.make_named_collection_extractor(cast(Hashable, type(value)))(value)
    if isinstance(value, tuple):
        if needs_value_extraction(value):
            return tuple(extract(v, pass_through_values=pass_through_values) for v in value)  # type: ignore[union-attr]  # value is always iterable
        else:
            return value
    if pass_through_values:
        return cast(common.PrimitiveValue, value)

    raise TypeError(f"Cannot extract numeric value from {type(value)}.")


ExtractedArgs: TypeAlias = tuple[MaybeNestedInTuple[common.PrimitiveValue]]
ExtractedKwargs: TypeAlias = dict[str, MaybeNestedInTuple[common.PrimitiveValue]]


# TODO(egparedes): memoize this function (and/or the one above) if TypeSpecs become hashable
def make_primitive_value_args_extractor(
    function: ts.FunctionType,
) -> Callable[..., tuple[ExtractedArgs, ExtractedKwargs]] | None:
    """
    Make a function to extract primitive values from arguments that need it.

    If no arguments need extraction, return `None`.

    The returned function has the signature `(*args, **kwargs) -> (args, kwargs)`,
    where `args` is a tuple of positional arguments and `kwargs` is a dictionary of
    keyword arguments containing the extracted primitive values where needed.

    This function only uses structural information of the type, primitive values may be
    passed as :ref:`ts.DeferredType`.
    """

    args_param = "args"
    kwargs_param = "kwargs"
    num_args_to_extract = 0
    num_kwargs_to_extract = 0
    extractor_exprs: dict[int | str, str] = {}

    for i, type_spec in enumerate(
        pos_args := [
            *function.pos_only_args,
            *function.pos_or_kw_args.values(),
        ]
    ):
        if type_info.needs_value_extraction(type_spec):
            num_args_to_extract += 1
            extractor_exprs[i] = named_collections.make_extractor_expr_from_type_spec(
                type_spec, f"{args_param}[{i}]"
            )
        else:
            extractor_exprs[i] = f"{args_param}[{i}]"

    for name, type_spec in function.kw_only_args.items():
        if type_info.needs_value_extraction(type_spec):
            num_kwargs_to_extract += 1
            extractor_exprs[name] = named_collections.make_extractor_expr_from_type_spec(
                type_spec, f"{kwargs_param}['{name}']"
            )
        else:
            extractor_exprs[name] = f"{kwargs_param}['{name}']"

    if num_args_to_extract + num_kwargs_to_extract:
        args_expr = (
            f"({str.join(', ', (extractor_exprs[i] for i, _ in enumerate(pos_args)))}, )"
            if num_args_to_extract
            else args_param
        )
        kwargs_expr = (
            _make_dict_expr({k: extractor_exprs[k] for k in function.kw_only_args})
            if num_kwargs_to_extract
            else kwargs_param
        )
        extractor_func_src = f"lambda *{args_param}, **{kwargs_param}: ({args_expr}, {kwargs_expr})"
        return eval(extractor_func_src)

    return None



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""GridTools storages utilities."""

from . import cartesian
from .cartesian import layout
from .cartesian.interface import empty, from_array, full, ones, zeros
from .cartesian.layout import from_name, register


__all__ = [
    "cartesian",
    "empty",
    "from_array",
    "from_name",
    "full",
    "layout",
    "ones",
    "register",
    "zeros",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import functools
import operator
import re
from typing import Literal, Optional

from gt4py import eve
from gt4py.next import common, utils
from gt4py.next.iterator import ir, ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.transforms import (
    fixed_point_transformation,
    inline_lambdas,
    inline_lifts,
    symbol_ref_utils,
)
from gt4py.next.iterator.type_system import (
    inference as itir_type_inference,
    type_specifications as it_ts,
)
from gt4py.next.type_system import type_info, type_specifications as ts


def _with_altered_iterator_element_type(
    type_: it_ts.IteratorType, new_el_type: ts.DataType
) -> it_ts.IteratorType:
    return it_ts.IteratorType(
        position_dims=type_.position_dims, defined_dims=type_.defined_dims, element_type=new_el_type
    )


def _with_altered_iterator_position_dims(
    type_: it_ts.IteratorType, new_position_dims: list[common.Dimension] | Literal["unknown"]
) -> it_ts.IteratorType:
    return it_ts.IteratorType(
        position_dims=new_position_dims,
        defined_dims=type_.defined_dims,
        element_type=type_.element_type,
    )


def _is_trivial_make_tuple_call(node: itir.Expr):
    """Return if node is a `make_tuple` call with all elements `SymRef`s, `Literal`s or tuples thereof."""
    if not cpm.is_call_to(node, "make_tuple"):
        return False
    if not all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args):
        return False
    return True


def _is_trivial_or_tuple_thereof_expr(node: itir.Node) -> bool:
    """
    Return `true` if the expr is a trivial expression (`SymRef` or `Literal`) or tuple thereof.

    Let forms with trivial body and args as well as `if` and `concat_where` calls with trivial
    branches are also considered trivial.

    >>> _is_trivial_or_tuple_thereof_expr(im.make_tuple("a", "b"))
    True
    >>> _is_trivial_or_tuple_thereof_expr(im.tuple_get(1, "a"))
    True
    >>> _is_trivial_or_tuple_thereof_expr(
    ...     im.let("t", im.make_tuple("a", "b"))(im.tuple_get(1, "t"))
    ... )
    True
    """
    if isinstance(node, (itir.SymRef, itir.Literal)):
        return True
    if cpm.is_call_to(node, "make_tuple"):
        return all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args)
    if cpm.is_call_to(node, "tuple_get"):
        return _is_trivial_or_tuple_thereof_expr(node.args[1])
    # This will duplicate the condition and increase the size of the tree, but this is probably
    # acceptable.
    if cpm.is_call_to(node, ("if_", "concat_where")):
        return all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args[1:])
    if cpm.is_let(node):
        return _is_trivial_or_tuple_thereof_expr(node.fun.expr) and all(
            _is_trivial_or_tuple_thereof_expr(arg) for arg in node.args
        )
    return False


def _flattened_as_fieldop_param_el_name(param: str, idx: int) -> str:
    prefix = "__ct_flat_el_"

    # keep the original param name, but skip prefix from previous flattenings
    if param.startswith(prefix):
        parent_idx, suffix = re.split(r"_(?!\d)", param[len(prefix) :], maxsplit=1)
        prefix = f"{prefix}{parent_idx}_"
    else:
        suffix = param

    return f"{prefix}{idx}_{suffix}"


# TODO(tehrengruber): Conceptually the structure of this pass makes sense: Visit depth first,
#  transform each node until no transformations apply anymore, whenever a node is to be transformed
#  go through all available transformation and apply them. However the final result here still
#  reads a little convoluted and is also different to how we write other transformations. We
#  should revisit the pattern here and try to find a more general mechanism.
@dataclasses.dataclass(frozen=True, kw_only=True)
class CollapseTuple(
    fixed_point_transformation.CombinedFixedPointTransform, eve.PreserveLocationVisitor
):
    """
    Simplifies `make_tuple`, `tuple_get` calls.

      - `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
      - `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
    """

    # TODO(tehrengruber): This Flag mechanism is a little low level. What we actually want
    #   is something like a pass manager, where for each pattern we have a corresponding
    #   transformation, etc.
    class Transformation(enum.Flag):
        #: `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
        COLLAPSE_MAKE_TUPLE_TUPLE_GET = enum.auto()
        #: `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
        COLLAPSE_TUPLE_GET_MAKE_TUPLE = enum.auto()
        #: `tuple_get(i, let(...)(make_tuple()))` -> `let(...)(tuple_get(i, make_tuple()))`
        PROPAGATE_TUPLE_GET = enum.auto()
        #: `{1, 2}` -> `(λ(_tuple_el_1, _tuple_el_2) → {_tuple_el_1, _tuple_el_2})(1, 2)`
        LETIFY_MAKE_TUPLE_ELEMENTS = enum.auto()
        #: `let(tup, {trivial_expr1, trivial_expr2})(foo(tup))`
        #:  -> `foo({trivial_expr1, trivial_expr2})`
        INLINE_TRIVIAL_MAKE_TUPLE = enum.auto()
        #:  Similar as `PROPAGATE_TO_IF_ON_TUPLES`, but propagates in the opposite direction, i.e.
        #:  into the tree, allowing removal of tuple expressions across `if_` calls without
        #:  increasing the size of the tree. This is particularly important for `if` statements
        #:  in the frontend, where outwards propagation can have devastating effects on the tree
        #:  size, without any gained optimization potential. For example
        #:  ```
        #:   complex_lambda(if cond1
        #:     if cond2
        #:       {...}
        #:     else:
        #:       {...}
        #:   else
        #:     {...})
        #:  ```
        #:  is problematic, since `PROPAGATE_TO_IF_ON_TUPLES` would propagate and hence duplicate
        #:  `complex_lambda` three times, while we only want to get rid of the tuple expressions
        #:  inside of the `if_`s.
        #:  Note that this transformation is not mutually exclusive to `PROPAGATE_TO_IF_ON_TUPLES`.
        PROPAGATE_TO_IF_ON_TUPLES_CPS = enum.auto()
        #: `(if cond then {1, 2} else {3, 4})[0]` -> `if cond then {1, 2}[0] else {3, 4}[0]`
        PROPAGATE_TO_IF_ON_TUPLES = enum.auto()
        #: `let((a, let(b, 1)(a_val)))(a)`-> `let(b, 1)(let(a, a_val)(a))`
        PROPAGATE_NESTED_LET = enum.auto()
        #: `let(a, 1)(a)` -> `1` or `let(a, b)(f(a))` -> `f(a)`
        INLINE_TRIVIAL_LET = enum.auto()
        #: `as_fieldop(λ(t) → ·t[0]+·t[1])({a, b})` -> `as_fieldop(λ(a, b) → ·a+·b)(a, b)`
        FLATTEN_AS_FIELDOP_ARGS = enum.auto()
        #: `let(a, b[1])(a)` -> `b[1]`
        INLINE_TRIVIAL_TUPLE_LET_VAR = enum.auto()

        @classmethod
        def all(self) -> CollapseTuple.Transformation:
            return functools.reduce(operator.or_, self.__members__.values())

    uids: utils.IDGeneratorPool
    enabled_transformations: Transformation = Transformation.all()  # noqa: RUF009 [function-call-in-dataclass-default-argument]

    REINFER_TYPES = True

    PRESERVED_ANNEX_ATTRS = ("type", "domain")

    @classmethod
    def apply(
        cls,
        node: itir.Node,
        *,
        remove_letified_make_tuple_elements: bool = True,
        offset_provider_type: Optional[common.OffsetProviderType] = None,
        within_stencil: Optional[bool] = None,
        # manually passing enabled transformations is mostly for allowing separate testing of the modes
        enabled_transformations: Optional[Transformation] = None,
        # allow sym references without a symbol declaration, mostly for testing
        allow_undeclared_symbols: bool = False,
        uids: utils.IDGeneratorPool,
    ) -> itir.Node:
        """
        Simplifies `make_tuple`, `tuple_get` calls.

        Arguments:
            node: The node to transform.

        Keyword arguments:
            remove_letified_make_tuple_elements: Run `InlineLambdas` as a post-processing step
                to remove left-overs from `LETIFY_MAKE_TUPLE_ELEMENTS` transformation.
                `(λ(_tuple_el_1, _tuple_el_2) → {_tuple_el_1, _tuple_el_2})(1, 2)` -> {1, 2}`
        """
        enabled_transformations = enabled_transformations or cls.enabled_transformations
        offset_provider_type = offset_provider_type or {}

        if isinstance(node, itir.Program):
            within_stencil = False
        assert within_stencil in [
            True,
            False,
        ], "Parameter 'within_stencil' mandatory if node is not a 'Program'."

        requires_types = False
        if enabled_transformations & (
            cls.Transformation.PROPAGATE_TO_IF_ON_TUPLES_CPS
            | cls.Transformation.FLATTEN_AS_FIELDOP_ARGS
        ):
            requires_types = True

        if requires_types:
            node = itir_type_inference.infer(
                node,
                offset_provider_type=offset_provider_type,
                allow_undeclared_symbols=allow_undeclared_symbols,
            )

        new_node = cls(
            enabled_transformations=enabled_transformations,
            uids=uids,
        ).visit(node, within_stencil=within_stencil)

        # inline to remove left-overs from LETIFY_MAKE_TUPLE_ELEMENTS. this is important
        # as otherwise two equal expressions containing a tuple will not be equal anymore
        # and the CSE pass can not remove them.
        # TODO(tehrengruber): test case for `scan(lambda carry: {1, 2})`
        #  (see solve_nonhydro_stencil_52_like_z_q_tup)
        if remove_letified_make_tuple_elements:
            new_node = inline_lambdas.InlineLambdas.apply(
                new_node, opcount_preserving=True, force_inline_lambda_args=False
            )

        return new_node

    def visit(self, node, **kwargs):
        if cpm.is_call_to(node, "as_fieldop"):
            kwargs = {**kwargs, "within_stencil": True}

        return super().visit(node, **kwargs)

    def transform_collapse_make_tuple_tuple_get(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "make_tuple") and all(
            cpm.is_call_to(arg, "tuple_get") for arg in node.args
        ):
            # `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
            assert len(node.args) > 0 and isinstance(node.args[0], itir.FunCall)
            first_expr = node.args[0].args[1]

            for i, v in enumerate(node.args):
                assert isinstance(v, itir.FunCall)
                assert isinstance(v.args[0], itir.Literal)
                if not (int(v.args[0].value) == i and ir_misc.is_equal(v.args[1], first_expr)):
                    # tuple argument differs, just continue with the rest of the tree
                    return None

            itir_type_inference.reinfer(first_expr)  # type is needed so reinfer on-demand
            assert isinstance(first_expr.type, (ts.TupleType, ts.DeferredType))
            if isinstance(first_expr.type, ts.TupleType) and len(first_expr.type.types) == len(
                node.args
            ):
                return first_expr
        return None

    def transform_collapse_tuple_get_make_tuple(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if (
            cpm.is_call_to(node, "tuple_get")
            and isinstance(node.args[0], itir.Literal)
            and cpm.is_call_to(node.args[1], "make_tuple")
        ):
            # `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
            assert not node.args[0].type or type_info.is_integer(node.args[0].type)
            make_tuple_call = node.args[1]
            idx = int(node.args[0].value)
            assert idx < len(make_tuple_call.args), (
                f"Index {idx} is out of bounds for tuple of size {len(make_tuple_call.args)}"
            )
            return node.args[1].args[idx]
        return None

    def transform_propagate_tuple_get(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "tuple_get") and isinstance(node.args[0], itir.Literal):
            # TODO(tehrengruber): extend to general symbols as long as the tail call in the let
            #   does not capture
            # `tuple_get(i, let(...)(make_tuple()))` -> `let(...)(tuple_get(i, make_tuple()))`
            idx, expr = node.args
            assert isinstance(idx, itir.Literal)
            if cpm.is_let(expr):
                return im.call(
                    im.lambda_(*expr.fun.params)(
                        self.fp_transform(im.tuple_get(idx.value, expr.fun.expr), **kwargs)
                    )
                )(*expr.args)
            elif cpm.is_call_to(expr, ("if_", "concat_where")):
                fun = expr.fun
                cond, true_branch, false_branch = expr.args
                return im.call(fun)(
                    cond,
                    self.fp_transform(im.tuple_get(idx.value, true_branch), **kwargs),
                    self.fp_transform(im.tuple_get(idx.value, false_branch), **kwargs),
                )
        return None

    def transform_letify_make_tuple_elements(
        self, node: itir.Node, **kwargs
    ) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "make_tuple"):
            # `make_tuple(expr1, expr1)`
            # -> `let((_tuple_el_1, expr1), (_tuple_el_2, expr2))(make_tuple(_tuple_el_1, _tuple_el_2))`
            bound_vars: dict[itir.Sym, itir.Expr] = {}
            new_args: list[itir.Expr] = []
            for arg in node.args:
                if cpm.is_call_to(node, "make_tuple") and not _is_trivial_make_tuple_call(node):
                    new_arg = im.ref(next(self.uids["__ct_el"]), arg.type)
                    self._preserve_annex(arg, new_arg)
                    new_args.append(new_arg)
                    bound_vars[im.sym(new_arg.id, arg.type)] = arg
                else:
                    new_args.append(arg)

            if bound_vars:
                return self.fp_transform(
                    im.let(*bound_vars.items())(im.call(node.fun)(*new_args)), **kwargs
                )
        return None

    def transform_inline_trivial_make_tuple(self, node: itir.Node, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            # `let(tup, make_tuple(trivial_expr1, trivial_expr2))(foo(tup))`
            #  -> `foo(make_tuple(trivial_expr1, trivial_expr2))`
            eligible_params = [_is_trivial_make_tuple_call(arg) for arg in node.args]
            if any(eligible_params):
                return self.visit(
                    inline_lambdas.inline_lambda(node, eligible_params=eligible_params), **kwargs
                )
        return None

    def transform_propagate_to_if_on_tuples(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if kwargs["within_stencil"]:
            # TODO(tehrengruber): This significantly increases the size of the tree. Skip transformation
            #  in local-view for now. Revisit.
            return None

        if isinstance(node, itir.FunCall) and not cpm.is_call_to(node, "if_"):
            # TODO(tehrengruber): Only inline if type of branch value is a tuple.
            # Examples:
            # `(if cond then {1, 2} else {3, 4})[0]` -> `if cond then {1, 2}[0] else {3, 4}[0]`
            # `let (b, if cond then {1, 2} else {3, 4})) b[0]`
            #  -> `if cond then let(b, {1, 2})(b[0]) else let(b, {3, 4})(b[0])`
            for i, arg in enumerate(node.args):
                if cpm.is_call_to(arg, "if_"):
                    cond, true_branch, false_branch = arg.args
                    new_true_branch = self.fp_transform(
                        ir_misc.with_altered_arg(node, i, true_branch), **kwargs
                    )
                    new_false_branch = self.fp_transform(
                        ir_misc.with_altered_arg(node, i, false_branch), **kwargs
                    )
                    return im.if_(cond, new_true_branch, new_false_branch)
        return None

    def transform_propagate_to_if_on_tuples_cps(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        # The basic idea of this transformation is to remove tuples across if-stmts by rewriting
        # the expression in continuation passing style, e.g. something like a tuple reordering
        # ```
        # let t = if True then {1, 2} else {3, 4} in
        #   {t[1], t[0]})
        # end
        # ```
        # is rewritten into:
        # ```
        # let cont = λ(el0, el1) → {el1, el0} in
        #  if True then cont(1, 2) else cont(3, 4)
        # end
        # ```
        # Note how the `make_tuple` call argument of the `if` disappears. Since lambda functions
        # are currently inlined (due to limitations of the domain inference) we will only
        # gain something compared `PROPAGATE_TO_IF_ON_TUPLES` if the continuation `cont` is trivial,
        # e.g. a `make_tuple` call like in the example. In that case we can inline the trivial
        # continuation and end up with an only moderately larger tree, e.g.
        # `if True then {2, 1} else {4, 3}`. The examples in the comments below all refer to this
        # tuple reordering example here.

        if not isinstance(node, itir.FunCall) or cpm.is_call_to(node, "if_"):
            return None

        # The first argument that is eligible also transforms all remaining args (They will be
        # part of the continuation which is recursively transformed).
        for i, arg in enumerate(node.args):
            if cpm.is_call_to(arg, "if_"):
                itir_type_inference.reinfer(arg)

                cond, true_branch, false_branch = arg.args  # e.g. `True`, `{1, 2}`, `{3, 4}`
                if not any(
                    isinstance(branch.type, ts.TupleType) for branch in [true_branch, false_branch]
                ):
                    continue
                tuple_type: ts.TupleType = true_branch.type  # type: ignore[assignment]  # type ensured above
                tuple_len = len(tuple_type.types)

                # build and simplify continuation, e.g. λ(el0, el1) → {el1, el0}
                itir_type_inference.reinfer(node)
                assert node.type
                f_type = ts.FunctionType(  # type of continuation in order to keep full type info
                    pos_only_args=tuple_type.types,
                    pos_or_kw_args={},
                    kw_only_args={},
                    returns=node.type,
                )
                f_params = [
                    im.sym(next(self.uids["__ct_el_cps"]), type_) for type_ in tuple_type.types
                ]
                f_args = [im.ref(param.id, param.type) for param in f_params]
                f_body = ir_misc.with_altered_arg(node, i, im.make_tuple(*f_args))

                # if the function is not trivial the transformation would create a larger tree
                # after inlining so we skip transformation this argument.
                if not _is_trivial_or_tuple_thereof_expr(f_body):
                    continue

                # simplify, e.g., inline trivial make_tuple args
                new_f_body = self.fp_transform(f_body, **kwargs)
                # if the continuation did not simplify there is nothing to gain. Skip
                # transformation of this argument.
                if new_f_body is f_body:
                    continue

                f = im.lambda_(*f_params)(new_f_body)

                # this is the symbol refering to the tuple value inside the two branches of the
                # if, e.g. a symbol refering to `{1, 2}` and `{3, 4}` respectively
                tuple_var = next(self.uids["__ct_tuple_cps"])
                # this is the symbol refering to our continuation, e.g. `cont` in our example.
                f_var = next(self.uids["__ct_cont"])
                new_branches = []
                for branch in [true_branch, false_branch]:
                    new_branch = im.let(tuple_var, branch)(
                        im.call(im.ref(f_var, f_type))(  # call to the continuation
                            *(
                                im.tuple_get(i, im.ref(tuple_var, branch.type))
                                for i in range(tuple_len)
                            )
                        )
                    )
                    new_branches.append(self.fp_transform(new_branch, **kwargs))

                # assemble everything together
                new_node = im.let(f_var, f)(im.if_(cond, *new_branches))
                new_node = inline_lambdas.inline_lambda(new_node, eligible_params=[True])
                assert cpm.is_call_to(new_node, "if_")
                new_node = im.if_(
                    cond, *(self.fp_transform(branch, **kwargs) for branch in new_node.args[1:])
                )
                return new_node

        return None

    def transform_propagate_nested_let(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            # `let((a, let(b, 1)(a_val)))(a)`-> `let(b, 1)(let(a, a_val)(a))`
            outer_vars: dict[itir.Sym, itir.Expr] = {}
            inner_vars: dict[itir.Sym, itir.Expr] = {}
            original_inner_expr = node.fun.expr

            # Here we store all references used in let args and the inner expression. Variable
            # is only populated below in case the transformation applies.
            reserved_refs = None

            for arg_sym, arg in zip(node.fun.params, node.args):
                assert arg_sym not in inner_vars
                if cpm.is_let(arg):
                    # TODO(tehrengruber): This is potentially expensive. Store the used symbol refs in the annex.
                    if not reserved_refs:
                        reserved_refs = symbol_ref_utils.collect_symbol_refs(
                            [*node.args, original_inner_expr]
                        )
                    rename_map: dict[
                        str, ir.SymRef
                    ] = {}  # mapping from symbol with a collision to its new (unique) name
                    for sym, val in zip(arg.fun.params, arg.args):
                        unique_sym = ir_misc.unique_symbol(
                            sym, [*(s.id for s in outer_vars.keys()), *reserved_refs]
                        )
                        if sym != unique_sym:  # name collision, rename symbol to unique_sym later
                            assert sym.id not in rename_map
                            rename_map[sym.id] = im.ref(unique_sym.id, sym.type)

                        outer_vars[unique_sym] = val

                    new_expr = arg.fun.expr
                    if rename_map:
                        new_expr = inline_lambdas.rename_symbols(new_expr, rename_map)

                    inner_vars[arg_sym] = new_expr
                else:
                    inner_vars[arg_sym] = arg
            if outer_vars:
                return self.fp_transform(
                    im.let(*outer_vars.items())(
                        self.fp_transform(
                            im.let(*inner_vars.items())(original_inner_expr), **kwargs
                        )
                    ),
                    **kwargs,
                )
        return None

    def transform_inline_trivial_let(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            if isinstance(node.fun.expr, itir.SymRef):
                # `let(a, 1)(a)` -> `1`
                for arg_sym, arg in zip(node.fun.params, node.args):
                    if isinstance(node.fun.expr, itir.SymRef) and node.fun.expr.id == arg_sym.id:
                        return arg
            if any(
                trivial_args := [isinstance(arg, (itir.SymRef, itir.Literal)) for arg in node.args]
            ):
                return inline_lambdas.inline_lambda(node, eligible_params=trivial_args)

        return None

    def transform_inline_trivial_tuple_let_var(self, node: ir.Node, **kwargs) -> Optional[ir.Node]:
        if cpm.is_let(node):
            if any(trivial_args := [_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args]):
                return inline_lambdas.inline_lambda(node, eligible_params=trivial_args)
        return None

    # TODO(tehrengruber): This is a transformation that should be executed before visiting the children. Then
    #  revisiting the body would not be needed.
    def transform_flatten_as_fieldop_args(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        # `as_fieldop(λ(t) → ·t[0]+·t[1])({a, b})` -> `as_fieldop(λ(a, b) → ·a+·b)(a, b)`
        if not cpm.is_applied_as_fieldop(node):
            return None

        for arg in node.args:
            itir_type_inference.reinfer(arg)

        if not any(isinstance(arg.type, ts.TupleType) for arg in node.args):
            return None

        node = ir_misc.canonicalize_as_fieldop(node)
        stencil, restore_scan = ir_misc.unwrap_scan(
            node.fun.args[0]  # type: ignore[attr-defined] # ensured by cpm.is_applied_as_fieldop
        )

        new_body = stencil.expr
        domain = node.fun.args[1] if len(node.fun.args) > 1 else None  # type: ignore[attr-defined] # ensured by cpm.is_applied_as_fieldop
        remapped_args: dict[
            itir.Sym, itir.Expr
        ] = {}  # contains the arguments that are remapped, e.g. `{a, b}`
        new_params: list[itir.Sym] = []
        new_args: list[itir.Expr] = []
        for param, arg in zip(stencil.params, node.args, strict=True):
            if isinstance(arg.type, ts.TupleType):
                assert isinstance(param.type, it_ts.IteratorType)
                ref_to_remapped_arg = im.ref(
                    f"__ct_flat_remapped_{len(remapped_args)}",
                    arg.type,
                )
                self._preserve_annex(arg, ref_to_remapped_arg)
                remapped_args[im.sym(ref_to_remapped_arg.id, arg.type)] = arg
                new_params_inner, lift_params = [], []
                assert isinstance(param.type.element_type, ts.TupleType)
                for i, type_ in enumerate(param.type.element_type.types):
                    assert isinstance(type_, ts.DataType)
                    new_param = im.sym(
                        _flattened_as_fieldop_param_el_name(param.id, i),
                        _with_altered_iterator_element_type(param.type, type_),
                    )
                    lift_params.append(
                        im.sym(
                            new_param.id,
                            _with_altered_iterator_position_dims(new_param.type, "unknown"),  # type: ignore[arg-type]  # always in IteratorType
                        )
                    )
                    new_params_inner.append(new_param)
                    new_args.append(im.tuple_get(i, ref_to_remapped_arg))

                # an iterator that substitutes the original (tuple) iterator, e.g. `t`. Built
                # from the new parameters which are the elements of `t`.
                param_substitute = im.lift(
                    im.lambda_(*lift_params)(
                        im.make_tuple(*[im.deref(im.ref(p.id, p.type)) for p in lift_params])
                    )
                )(*[im.ref(p.id, p.type) for p in new_params_inner])

                new_body = im.let(param.id, param_substitute)(new_body)
                # note: the lift is trivial so inlining it is not an issue with respect to tree size
                new_body = inline_lambdas.inline_lambda(new_body, force_inline_lift_args=True)

                new_params.extend(new_params_inner)
            else:
                new_params.append(param)
                new_args.append(arg)

        # remove lifts again
        new_body = inline_lifts.InlineLifts(
            flags=inline_lifts.InlineLifts.Flag.INLINE_DEREF_LIFT
            | inline_lifts.InlineLifts.Flag.PROPAGATE_SHIFT
        ).visit(new_body)
        new_body = self.visit(new_body, **kwargs)
        new_stencil = restore_scan(im.lambda_(*new_params)(new_body))

        return im.let(*remapped_args.items())(im.as_fieldop(new_stencil, domain)(*new_args))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import warnings
from typing import Any, Callable, Mapping, Optional, TypeAlias, Union

import dace
from dace import (
    properties as dace_properties,
    subsets as dace_subsets,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


MapPromotionCallBack: TypeAlias = Callable[
    [dace.SDFGState, dace.SDFG, dace_nodes.MapExit, dace_nodes.MapEntry, list[str]], bool
]
"""Callback for the `MapPromoter`.

After the `MapPromoter` has checked that the map would be promoted, the
callback is called. If the function returns `True` then transformation will
perform the promotion and in case of `False` the transformation will not
perform the promotion.
This allows user code to steer the promotion.

Args:
    state: The `SDFGState` on which the transformation operate.
    sdfg: The `SDFG` on which the transformation operate.
    first_map_exit: The `MapExit` node of the first Map, the one that is promoted.
    second_map_entry: The `MapEntry` node of the second Map.
    missing_map_parameters: The list of Map parameters that will be added to the
        first map.
"""


@dace_properties.make_properties
class MapPromoter(dace_transformation.SingleStateTransformation):
    """Promotes a Map such that it can be fused together.

    The transformation essentially matches `MapExit -> (Intermediate) -> MapEntry`
    and modifies the first Map such that it can be fused. The essentially checks
    if the parameter of the first Map are a subset of the second Map, in case
    they match they are checked for equality. If this is the case the
    transformation will add the missing parameters to the first map.
    It is possible to influence what kind of dimensions can be added.

    As an example, the transformation will turn the following code:
    ```python
    for i in dace.map[0:N]:
        a[i] = foo(i, ...)
    for i, j in dace.map[0:N, 0:M]:
        b[i, j] = bar(a[i], i, j, ...)
    ```
    into
    ```python
    for i, j in dace.map[0:N, 0:M]:
        a[i] = foo(i, ...)
    for i, j in dace.map[0:N, 0:M]:
        b[i, j] = bar(a[i], i, j, ...)
    ```
    which can be fused together by map fusion.

    Args:
        only_inner_maps: Only match Maps that are internal, i.e. inside another Map.
        only_toplevel_maps: Only consider Maps that are at the top.
        promote_vertical: If `True` promote vertical dimensions, i.e. add them
            to the map to promote; `True` by default.
        promote_local: If `True` promote local dimensions, i.e. add them to the
            map to promote; `True` by default.
        promote_horizontal: If `True` promote horizontal dimensions, i.e. add
            them to the map to promote; `False` by default.
        promote_everything: Do not impose any restriction on what to promote. The only
            reasonable value is `True` or `False`.
        fuse_after_promotion: If `True`, the default, then fuse the two maps together
            immediately after promotion, i.e. inside `apply()`.
        promotion_callback: A callback function, see `MapPromotionCallBack`, that
            can be used to steer the promotion.

    Note:
        - The transformation will always promote the top Map never the lower Map.
        - This ignores tiling.

    Todo:
        Promotion in vertical direction, i.e. adding a vertical dimension is
        most likely all the time good, because it will create independent nodes
        that `LoopBlocking` can move out of the inner loop. However, promotion
        in horizontal, i.e. adding horizontal dimensions, is not necessarily
        good. Empirical observations have shown that it is most likely, but
        we should add a criteria to prevent the promotion in certain cases.
    """

    only_toplevel_maps = dace_properties.Property(
        dtype=bool,
        default=False,
        allow_none=False,
        desc="Only perform fusing if the Maps are on the top level.",
    )
    only_inner_maps = dace_properties.Property(
        dtype=bool,
        default=False,
        allow_none=False,
        desc="Only perform fusing if the Maps are inner Maps, i.e. does not have top level scope.",
    )
    promote_vertical = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If `True` promote vertical dimensions.",
    )
    promote_local = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If `True` promote local dimensions.",
    )
    promote_horizontal = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If `True` promote horizontal dimensions.",
    )
    promote_everything = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If `True` perform any promotion. Takes precedence over all other selectors.",
    )
    fuse_after_promotion = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If `True` fuse the maps together immediately after promotion.",
    )

    _promotion_callback: Optional[MapPromotionCallBack]

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: dict[dace.SDFG, set[str]]

    # Pattern Matching
    exit_first_map = dace_transformation.PatternNode(dace_nodes.MapExit)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    entry_second_map = dace_transformation.PatternNode(dace_nodes.MapEntry)

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.exit_first_map, cls.access_node, cls.entry_second_map
            )
        ]

    def __init__(
        self,
        only_inner_maps: Optional[bool] = None,
        only_toplevel_maps: Optional[bool] = None,
        promote_local: Optional[bool] = None,
        promote_vertical: Optional[bool] = None,
        promote_horizontal: Optional[bool] = None,
        promote_everything: Optional[bool] = None,
        fuse_after_promotion: Optional[bool] = None,
        promotion_callback: Optional[MapPromotionCallBack] = None,
        *args: Any,
        single_use_data: dict[dace.SDFG, set[str]],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if fuse_after_promotion is not None:
            self.fuse_after_promotion = fuse_after_promotion
        if only_inner_maps is not None:
            self.only_inner_maps = only_inner_maps
        if only_toplevel_maps is not None:
            self.only_toplevel_maps = only_toplevel_maps
        if promote_local is not None:
            self.promote_local = promote_local
        if promote_vertical is not None:
            self.promote_vertical = promote_vertical
        if promote_horizontal is not None:
            self.promote_horizontal = promote_horizontal
        if promote_everything is not None:
            self.promote_everything = promote_everything
            self.promote_horizontal = False
            self.promote_vertical = False
            self.promote_local = False
        if only_inner_maps and only_toplevel_maps:
            raise ValueError("You specified both `only_inner_maps` and `only_toplevel_maps`.")
        if not (
            self.promote_local
            or self.promote_vertical
            or self.promote_horizontal
            or self.promote_everything
        ):
            raise ValueError(
                "You must select at least one class of dimension that should be promoted."
            )

        self._promotion_callback = None
        if promotion_callback is not None:
            self._promotion_callback = promotion_callback

        self._single_use_data = single_use_data

        # This flag is only for testing. It allows to bypass the check if the
        #  Maps can be fused after promotion.
        self._bypass_fusion_test = False

        if not self.fuse_after_promotion:
            warnings.warn(
                "Created a `MapPromoter` that does not fuse immediately, which might lead to borderline invalid SDFGs.",
                stacklevel=1,
            )

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        first_map_exit: dace_nodes.MapExit = self.exit_first_map
        second_map_entry: dace_nodes.MapEntry = self.entry_second_map

        if self.only_inner_maps or self.only_toplevel_maps:
            scope_dict: Mapping[dace_nodes.Node, Union[dace_nodes.Node, None]] = graph.scope_dict()
            if self.only_inner_maps and (scope_dict[second_map_entry] is None):
                return False
            if self.only_toplevel_maps and (scope_dict[second_map_entry] is not None):
                return False

        # Test if the map ranges and parameter are compatible with each other
        missing_map_parameters: list[str] | None = self._missing_map_params(
            map_to_promote=first_map_exit.map,
            source_map=second_map_entry.map,
            be_strict=True,
        )
        if not missing_map_parameters:
            return False

        # We now know which dimensions we have to add to the promotee map.
        #  Now we must test if we are also allowed to make that promotion in the first place.
        if not self.promote_everything:
            allowed_missing_dimension_suffixes: list[str] = []
            if self.promote_local:
                allowed_missing_dimension_suffixes.append("_gtx_localdim")
            if self.promote_vertical:
                allowed_missing_dimension_suffixes.append("_gtx_vertical")
            if self.promote_horizontal:
                allowed_missing_dimension_suffixes.append("_gtx_horizontal")
            if not allowed_missing_dimension_suffixes:
                return False
            for missing_map_param in missing_map_parameters:
                # Check if all missing parameter match a specified pattern. Note
                #  unknown iteration parameter, such as `__hansi_meier` will be
                #  rejected and can not be promoted.
                if not any(
                    missing_map_param.endswith(dim_identifier)
                    for dim_identifier in allowed_missing_dimension_suffixes
                ):
                    return False

        # It might be that the ranges that we add to the first Map from the second Map
        #  are empty. Which means that the second Map is never executed and after
        #  promotion, the first Map that was executed before would no longer be run.
        #  To prevent that we require that the number of iterations the second Map
        #  performs, at compile time, is larger than zero.
        second_map_iterations: Any = second_map_entry.map.range.num_elements()
        if str(second_map_iterations).isdigit():
            second_map_iterations = int(str(second_map_iterations))
            if second_map_iterations <= 0:
                return False
        elif hasattr(second_map_iterations, "free_symbols"):
            # According to [issue#2095](https://github.com/spcl/dace/issues/2095) DaCe is quite
            #  liberal concerning the positivity assumption, but in GT4Py this is not possible.
            second_map_iterations = second_map_iterations.subs(
                (
                    (sym, dace.symbol(sym.name, nonnegative=False))
                    for sym in list(second_map_iterations.free_symbols)
                )
            )
            if (second_map_iterations > 0) != True:  # noqa: E712 [true-false-comparison]  # SymPy fuzzy bools.
                return False
        else:
            warnings.warn(
                "Was unable to determine if the second Map ({second_map_entry}) is executed.",
                stacklevel=0,
            )
            return False

        if self._bypass_fusion_test:
            pass
        elif not self._test_if_promoted_maps_can_be_fused(graph, sdfg):
            return False

        if self._promotion_callback is not None:
            if not self._promotion_callback(
                graph, sdfg, first_map_exit, second_map_entry, missing_map_parameters
            ):
                return False

        return True

    def _promote_first_map(
        self,
        first_map_exit: dace_nodes.MapExit,
        second_map_entry: dace_nodes.MapEntry,
    ) -> None:
        first_map_exit.map.params = copy.deepcopy(second_map_entry.map.params)
        first_map_exit.map.range = copy.deepcopy(second_map_entry.map.range)

    def apply(self, graph: Union[dace.SDFGState, dace.SDFG], sdfg: dace.SDFG) -> None:
        first_map_exit: dace_nodes.MapExit = self.exit_first_map
        access_node: dace_nodes.AccessNode = self.access_node
        second_map_entry: dace_nodes.MapEntry = self.entry_second_map

        # Now promote the second map such that it maps the first map.
        self._promote_first_map(first_map_exit, second_map_entry)

        if self.fuse_after_promotion:
            # Unlike in the `can_be_applied()` function we do specify here the right
            #  parameter, i.e. ensure that the intermediate can be removed. Because
            #  we can not pass the single use data, this will lead to a scan of the
            #  SDFG. However, we have to do it that way to get the desired result.
            gtx_transformations.MapFusionVertical.apply_to(
                sdfg=sdfg,
                options={
                    "only_inner_maps": self.only_inner_maps,
                    "only_toplevel_maps": self.only_toplevel_maps,
                    "require_exclusive_intermediates": True,
                    "require_all_intermediates": True,
                },
                # This will not run `MapFusionVertical.can_be_applied()`, thus we scan the
                #  SDFG only once instead of twice for every intermediate.
                verify=False,
                first_map_exit=first_map_exit,
                array=access_node,
                second_map_entry=second_map_entry,
            )

    def _missing_map_params(
        self,
        map_to_promote: dace_nodes.Map,
        source_map: dace_nodes.Map,
        be_strict: bool = True,
    ) -> list[str] | None:
        """Returns the parameter that are missing in the map that should be promoted.

        The returned sequence is empty if they are already have the same parameters.
        The function will return `None` is promoting is not possible.
        By setting `be_strict` to `False` the function will only check the names.

        Args:
            map_to_promote: The map that should be promoted.
            source_map: The map acting as template.
            be_strict: Ensure that the ranges that are already there are correct.
        """
        source_params_set: set[str] = set(source_map.params)
        curr_params_set: set[str] = set(map_to_promote.params)

        # The promotion can only work if the source map's parameters
        #  if a superset of the ones the map that should be promoted is.
        if not source_params_set.issuperset(curr_params_set):
            return None

        if be_strict:
            # Check if the parameters that are already in the map to promote have
            #  the same range as in the source map.
            source_ranges: dace_subsets.Range = source_map.range
            curr_ranges: dace_subsets.Range = map_to_promote.range
            curr_param_to_idx: dict[str, int] = {p: i for i, p in enumerate(map_to_promote.params)}
            source_param_to_idx: dict[str, int] = {p: i for i, p in enumerate(source_map.params)}
            for param_to_check in curr_params_set:
                curr_range = curr_ranges[curr_param_to_idx[param_to_check]]
                source_range = source_ranges[source_param_to_idx[param_to_check]]
                # TODO(phimuell): Use simplify?
                if curr_range != source_range:
                    return None
        return list(source_params_set - curr_params_set)

    def _test_if_promoted_maps_can_be_fused(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> bool:
        """This function checks if the promoted maps can be fused by map fusion.

        This function assumes that `super().self.can_be_applied()` returned `True`.

        Args:
            state: The state in which we operate.
            sdfg: The SDFG we process.
        """

        first_map_exit: dace_nodes.MapExit = self.exit_first_map
        access_node: dace_nodes.AccessNode = self.access_node
        second_map_entry: dace_nodes.MapEntry = self.entry_second_map

        # Since it is not possible to pass `self._single_use_data` to the Map fusion
        #  transformation, Map fusion would have to scan the SDFG on its own to
        #  figuring out if something is single use data or not. In order to avoid
        #  that we will now do the check ourselves. We not only check if the data is
        #  single use, but if it is only used (in this state) by the second Map. See
        #  bellow for why this is important.
        single_use_data = self._single_use_data[sdfg]
        for oedge in state.out_edges(first_map_exit):
            onode = oedge.dst
            if not isinstance(onode, dace_nodes.AccessNode):
                return False
            if not onode.desc(sdfg).transient:
                return False
            if onode.data not in single_use_data:
                return False
            if any(e.dst is not second_map_entry for e in state.out_edges(onode)):
                return False

        # Since we force a promotion of the map we have to store the old parameters
        #  of the map such that we can later restore them.
        first_map = first_map_exit.map
        org_first_map_params = copy.deepcopy(first_map.params)
        org_first_map_ranges = copy.deepcopy(first_map.range)

        try:
            # This will lead to a promotion of the map, this is needed that
            #  Map fusion can actually inspect them.
            self._promote_first_map(first_map_exit, second_map_entry)

            # Technically the promotion creates an invalid SDFG. Going back to the example
            #  in the doc string, after the promotion, but before the fusion, `a[i]` is
            #  written `M` times. This is not an issue per se, since each time the same
            #  value is written and after the fusion `a` will disappear, which ensures
            #  a valid SDFG. However, the removal of `a`, is only possible if the Map
            #  fusion classifies `a` as an "exclusive intermediate" (see `MapFusionVertical`).
            #  This is the reason why we have to set `require_all_intermediates`, to
            #  ensure that there are no "pure outputs" of the first Map.
            #  This also means that we would also have to specify `require_exclusive_intermediates`,
            #  because "shared intermediates" are not removed. But the test above has
            #  ensured that this is the case. As an optimization, to avoid that the
            #  Map fusion transformation scans the SDFG, we will specify `assume_always_shared`.
            #  This _is_ wrong, but since we only want to test if the Maps can be fused
            #  and this value has no influence on that outcome, we can specify it here.
            # TODO(phimuell): Once `single_use_data` can be passed, remove the optimization
            #   with `assume_always_shared`.
            if not gtx_transformations.MapFusionVertical.can_be_applied_to(
                sdfg=sdfg,
                options={
                    "only_inner_maps": self.only_inner_maps,
                    "only_toplevel_maps": self.only_toplevel_maps,
                    "require_all_intermediates": True,
                    "assume_always_shared": True,
                },
                first_map_exit=first_map_exit,
                array=access_node,
                second_map_entry=second_map_entry,
            ):
                return False

        finally:
            # Restore the parameters of the map that we promoted before.
            first_map.params = org_first_map_params
            first_map.range = org_first_map_ranges

        return True



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import collections.abc
import functools
import math
import numbers
from typing import Literal, Optional, Sequence, Tuple, Union, cast

import numpy as np
import numpy.typing as npt
from numpy.typing import DTypeLike

from gt4py._core import definitions as core_defs
from gt4py.eve.extended_typing import ArrayInterface, CUDAArrayInterface
from gt4py.storage import allocators


try:
    import cupy as cp
except ImportError:
    cp = None


FieldLike = Union["cp.ndarray", np.ndarray, ArrayInterface, CUDAArrayInterface]

_CPUBufferAllocator = allocators.NDArrayBufferAllocator(
    device_type=core_defs.DeviceType.CPU, array_utils=allocators.numpy_array_utils
)

_GPUBufferAllocator: Optional[allocators.NDArrayBufferAllocator] = None
if cp:
    assert isinstance(allocators.cupy_array_utils, allocators.ArrayUtils)

    if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.CUDA:
        _GPUBufferAllocator = allocators.NDArrayBufferAllocator(
            device_type=core_defs.DeviceType.CUDA,
            array_utils=allocators.cupy_array_utils,
        )
    elif core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
        _GPUBufferAllocator = allocators.NDArrayBufferAllocator(
            device_type=core_defs.DeviceType.ROCM,
            array_utils=allocators.cupy_array_utils,
        )
    else:
        raise ValueError("CuPy is available but no suitable device was found.")


def _idx_from_order(order):
    return list(np.argsort(order))


def _compute_padded_shape(shape, items_per_alignment, order_idx):
    padded_shape = list(shape)
    if len(order_idx) > 0:
        padded_shape[order_idx[-1]] = int(
            math.ceil(padded_shape[order_idx[-1]] / items_per_alignment) * items_per_alignment
        )
    return padded_shape


def _strides_from_padded_shape(padded_size, order_idx, itemsize):
    stride_accumulator = 1
    strides = [0] * len(padded_size)
    for idx in reversed(order_idx):
        strides[idx] = stride_accumulator * itemsize
        stride_accumulator = stride_accumulator * padded_size[idx]
    return list(strides)


def dimensions_to_mask(dimensions: Tuple[str, ...]) -> Tuple[bool, ...]:
    ndata_dims = sum(d.isdigit() for d in dimensions)
    mask = [(d in dimensions) for d in "IJK"] + [True for _ in range(ndata_dims)]
    return tuple(mask)


def normalize_storage_spec(
    aligned_index: Optional[Sequence[int]],
    shape: Sequence[int],
    dtype: DTypeLike,
    dimensions: Optional[Sequence[str]],
) -> Tuple[Sequence[int], Sequence[int], np.dtype, Tuple[str, ...]]:
    """Normalize the fields of the storage spec in a homogeneous representation.

    Returns
    -------
    tuple(aligned_index, shape, dtype, mask)
        The output tuple fields verify the following semantics:
            - aligned_index: tuple of ints with default origin values for the non-masked dimensions
            - shape: tuple of ints with shape values for the non-masked dimensions
            - dtype: scalar numpy.dtype (non-structured and without subarrays)
            - backend: backend identifier string (numpy, gt:cpu_kfirst, gt:gpu, ...)
            - dimensions: a tuple of dimension identifier strings
    """
    if dimensions is None:
        dimensions = (
            list("IJK"[: len(shape)])
            if len(shape) <= 3
            else list("IJK") + [str(d) for d in range(len(shape) - 3)]
        )

    if aligned_index is None:
        aligned_index = [0] * len(shape)

    dimensions = tuple(getattr(d, "__gt_axis_name__", d) for d in dimensions)
    if not all(isinstance(d, str) and (d.isdigit() or d in "IJK") for d in dimensions):
        raise ValueError(f"Invalid dimensions definition: '{dimensions}'")
    else:
        dimensions = tuple(str(d) for d in dimensions)
    if shape is not None:
        if not (
            isinstance(shape, collections.abc.Sequence)
            and all(isinstance(s, numbers.Integral) for s in shape)
        ):
            raise TypeError("shape must be an iterable of ints.")
        if len(shape) != len(dimensions):
            raise ValueError(
                f"Dimensions ({dimensions}) and shape ({shape}) have non-matching sizes."
                f"len(shape)(={len(shape)}) must be equal to len(dimensions)(={len(dimensions)})."
            )

        else:
            shape = tuple(shape)

        if any(i <= 0 for i in shape):
            raise ValueError(f"shape ({shape}) contains non-positive value.")
    else:
        raise TypeError("shape must be an iterable of ints.")

    if aligned_index is not None:
        if not (
            isinstance(aligned_index, collections.abc.Sequence)
            and all(isinstance(i, numbers.Integral) for i in aligned_index)
        ):
            raise TypeError("aligned_index must be an iterable of ints.")
        if len(aligned_index) != len(shape):
            raise ValueError(
                f"Shape ({shape}) and aligned_index ({aligned_index}) have non-matching sizes."
                f"len(aligned_index)(={len(aligned_index)}) must be equal to len(shape)(={len(shape)})."
            )

        aligned_index = tuple(aligned_index)

        if any(i < 0 for i in aligned_index):
            raise ValueError("aligned_index ({}) contains negative value.".format(aligned_index))
    else:
        raise TypeError("aligned_index must be an iterable of ints.")

    dtype = np.dtype(dtype)
    if dtype.shape:
        # Subarray dtype
        sub_dtype, sub_shape = cast(Tuple[np.dtype, Tuple[int, ...]], dtype.subdtype)
        aligned_index = (*aligned_index, *((0,) * dtype.ndim))
        shape = (*shape, *sub_shape)
        dimensions = (*dimensions, *(str(d) for d in range(dtype.ndim)))
        dtype = sub_dtype

    return aligned_index, shape, dtype, dimensions


def cpu_copy(array: Union[np.ndarray, "cp.ndarray"]) -> np.ndarray:
    if cp is not None:
        # it's not clear from the documentation if cp.asnumpy guarantees a copy.
        # worst case, this copies twice.
        return np.array(cp.asnumpy(array))
    return np.array(array)


def asarray(
    array: FieldLike, *, device: Literal["cpu", "gpu", None] = None
) -> np.ndarray | cp.ndarray:
    if hasattr(array, "ndarray"):
        # extract the buffer from a gt4py.next.Field
        # TODO(havogt): probably `Field` should provide the array interface methods when applicable
        array = array.ndarray

    xp = None
    if device == "cpu":
        xp = np
    elif device == "gpu":
        assert cp is not None
        xp = cp
    elif not device:
        if hasattr(array, "__dlpack_device__"):
            kind, _ = array.__dlpack_device__()
            if kind in [core_defs.DeviceType.CPU]:
                xp = np
            elif kind in [
                core_defs.DeviceType.CUDA,
                core_defs.DeviceType.ROCM,
            ]:
                if cp is None:
                    raise RuntimeError("CuPy is required for GPU arrays")
                xp = cp
        elif hasattr(array, "__cuda_array_interface__"):
            if cp is None:
                raise RuntimeError("CuPy is required for GPU arrays")
            xp = cp
        elif hasattr(array, "__array_interface__") or hasattr(array, "__array__"):
            xp = np

    if xp:
        return xp.asarray(array)

    if device is not None:
        raise ValueError(f"Invalid device: {device!s}")

    raise TypeError(f"Cannot convert {type(array)} to ndarray")


def get_dims(obj: Union[core_defs.GTDimsInterface, npt.NDArray]) -> Optional[Tuple[str, ...]]:
    dims = getattr(obj, "__gt_dims__", None)
    if dims is None:
        return dims
    return tuple(str(d) for d in dims)


def get_origin(obj: Union[core_defs.GTDimsInterface, npt.NDArray]) -> Optional[Tuple[int, ...]]:
    origin = getattr(obj, "__gt_origin__", None)
    if origin is None:
        return origin
    return tuple(int(o) for o in origin)


def allocate_cpu(
    shape: Sequence[int],
    layout_map: allocators.BufferLayoutMap,
    dtype: DTypeLike,
    alignment_bytes: int,
    aligned_index: Optional[Sequence[int]],
) -> Tuple[allocators._NDBuffer, np.ndarray]:
    device = core_defs.Device(core_defs.DeviceType.CPU, 0)
    buffer = _CPUBufferAllocator.allocate(
        shape,
        core_defs.dtype(dtype),
        device_id=device.device_id,
        layout_map=layout_map,
        byte_alignment=alignment_bytes,
        aligned_index=aligned_index,
    )
    return buffer.buffer, cast(np.ndarray, buffer.ndarray)


def _allocate_gpu(
    shape: Sequence[int],
    layout_map: allocators.BufferLayoutMap,
    dtype: DTypeLike,
    alignment_bytes: int,
    aligned_index: Optional[Sequence[int]],
) -> Tuple["cp.ndarray", "cp.ndarray"]:
    assert cp is not None
    assert _GPUBufferAllocator is not None, "GPU allocation library or device not found"
    if core_defs.CUPY_DEVICE_TYPE is None:
        raise ValueError("CUPY_DEVICE_TYPE detection failed.")
    device = core_defs.Device(  # type: ignore[type-var]
        core_defs.CUPY_DEVICE_TYPE, 0
    )
    buffer = _GPUBufferAllocator.allocate(
        shape,
        core_defs.dtype(dtype),
        device_id=device.device_id,
        layout_map=layout_map,
        byte_alignment=alignment_bytes,
        aligned_index=aligned_index,
    )

    buffer_ndarray = cast("cp.ndarray", buffer.ndarray)

    return buffer.buffer, buffer_ndarray


allocate_gpu = _allocate_gpu

if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:

    class CUDAArrayInterfaceNDArray(cp.ndarray):
        def __new__(cls, input_array: "cp.ndarray") -> CUDAArrayInterfaceNDArray:
            return (
                input_array
                if isinstance(input_array, CUDAArrayInterfaceNDArray)
                else cp.asarray(input_array).view(cls)
            )

        @property
        def __cuda_array_interface__(self) -> dict:
            return {
                "shape": self.shape,
                "typestr": self.dtype.descr[0][1],
                "descr": self.dtype.descr,
                "stream": 1,
                "version": 3,
                "strides": self.strides,
                "data": (self.data.ptr, False),
            }

        __hip_array_interface__ = __cuda_array_interface__

    @functools.wraps(_allocate_gpu)
    def _allocate_gpu_rocm(
        shape: Sequence[int],
        layout_map: allocators.BufferLayoutMap,
        dtype: DTypeLike,
        alignment_bytes: int,
        aligned_index: Optional[Sequence[int]],
    ) -> Tuple["cp.ndarray", "cp.ndarray"]:
        buffer, ndarray = _allocate_gpu(shape, layout_map, dtype, alignment_bytes, aligned_index)
        return buffer, CUDAArrayInterfaceNDArray(ndarray)

    allocate_gpu = _allocate_gpu_rocm



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""The GT4Py specific simplification pass."""

import collections
import copy
import uuid
import warnings
from typing import Any, Iterable, Optional, TypeAlias

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_subsets,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes, utils as dace_sdutils
from dace.transformation import (
    dataflow as dace_dataflow,
    interstate as dace_interstate,
    pass_pipeline as dace_ppl,
    passes as dace_passes,
)

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_simplify(
    sdfg: dace.SDFG,
    validate: bool = True,
    validate_all: bool = False,
    skip: Optional[Iterable[str]] = None,
) -> Optional[dict[str, Any]]:
    """Performs simplifications on the SDFG in place.

    Instead of calling `sdfg.simplify()` directly, you should use this function,
    as it is specially tuned for GridTool based SDFGs.

    This function runs the DaCe simplification pass, but the following passes are
    replaced:
    - `InlineSDFGs`: Instead `gt_inline_nested_sdfg()` will be called.
    - `FuseStates`: The normal DaCe transformation is still run, but after the DaCe
        simplify pass has ended the function will run `GT4PyStateFusion`.

    Further, the function will run the following passes in addition to DaCe simplify:
    - `SingleStateGlobalSelfCopyElimination`: Special copy pattern that in the context
        of GT4Py based SDFG behaves as a no op, i.e. `(G) -> (T) -> (G)`.
    - `SingleStateGlobalDirectSelfCopyElimination`: Special copy pattern of the form
        `(G) -> (G)` which can always be eliminated.
    - `MultiStateGlobalSelfCopyElimination`: Very similar to
        `SingleStateGlobalSelfCopyElimination`, with the exception that the write to
        `T`, i.e. `(G) -> (T)` and the write back to `G`, i.e. `(T) -> (G)` might be
        in different states.
    - `CopyChainRemover`: Which removes some chains that are introduced by the
        `concat_where` built-in function.
    - `GT4PyDeadDataflowElimination`: Run `gt_eliminate_dead_dataflow()` on the SDFG,
        which removes more dead dataflow than the native DaCe version.
    - `MapToCopy`: Called to remove some slices.
    - `canonicalize_memlet_trees()`: A free function that will canonicalize all Memlets.
        It is run after the inlining and can not be disabled.
    - `TrivialTaskletElimination`: Removing trivial copies.

    Furthermore, by default, or if `None` is passed for `skip` the passes listed in
    `GT_SIMPLIFY_DEFAULT_SKIP_SET` will be skipped.

    Args:
        sdfg: The SDFG to optimize.
        validate: Perform validation after the pass has run.
        validate_all: Perform extensive validation.
        skip: List of simplify passes that should not be applied, defaults
            to `GT_SIMPLIFY_DEFAULT_SKIP_SET`.

    Note:
        Currently DaCe does not provide a way to inject or exchange sub passes in
        simplify. The custom inline pass is run at the beginning and the array
        elimination at the end. The whole process is run inside a loop that ensures
        that `gt_simplify()` results in a fix point.
    """
    # Ensure that `skip` is a `set`
    skip = gtx_transformations.constants.GT_SIMPLIFY_DEFAULT_SKIP_SET if skip is None else set(skip)

    result: Optional[dict[str, Any]] = None

    at_least_one_xtrans_run = True
    starvation_protection = 30
    while at_least_one_xtrans_run:
        at_least_one_xtrans_run = False

        if starvation_protection == 0:
            raise ValueError("Simplify did not converge.")
        starvation_protection -= 1

        # NOTE: See comment in `gt_inline_nested_sdfg()` for more.
        sdfg.reset_cfg_list()

        if "InlineSDFGs" not in skip:
            inline_res = gt_inline_nested_sdfg(
                sdfg=sdfg,
                permissive=False,
                validate=False,
                validate_all=validate_all,
            )
            if inline_res is not None:
                at_least_one_xtrans_run = True
                result = result or {}
                result.update(inline_res)

        # Ensure that we have canonical Memelts.
        canoncialize_memlet_result = dace_sdutils.canonicalize_memlet_trees(sdfg)
        if canoncialize_memlet_result:
            result = result or {}
            if "canonicalize_memlet_trees" not in result:
                result["canonicalize_memlet_trees"] = 0
            result["canonicalize_memlet_trees"] += canoncialize_memlet_result

        simplify_res = dace_passes.SimplifyPass(
            validate=False,
            validate_all=validate_all,
            verbose=False,
            skip=(skip | {"InlineSDFGs"}),
        ).apply_pass(sdfg, {})

        if simplify_res is not None:
            at_least_one_xtrans_run = True
            result = result or {}
            result.update(simplify_res)

        # Note that it is not nice that we run the state fusion twice, but to be fully
        #  effective there are some preparatory transformations that are run in DaCe
        #  simplify. So the GT4Py transformation is more like a clean up to handle
        #  the parts DaCe is not able to do.
        if "FuseStates" not in skip:
            fuse_state_res = sdfg.apply_transformations_repeated(
                [gtx_transformations.GT4PyStateFusion],
                validate=False,
                validate_all=validate_all,
            )
            if fuse_state_res:
                at_least_one_xtrans_run = True
                result = result or {}
                if "FuseStates" not in result:
                    result["FuseStates"] = 0
                result["FuseStates"] += fuse_state_res

        if "MapToCopy" not in skip:
            find_single_use_data = dace_transformation.passes.analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
            removed_maps = sdfg.apply_transformations_once_everywhere(
                gtx_transformations.MapToCopy(
                    single_use_data=single_use_data,
                ),
                validate=False,
                validate_all=validate_all,
            )
            if removed_maps:
                at_least_one_xtrans_run = True
                result = result or {}
                if "MapToCopy" not in result:
                    result["MapToCopy"] = 0
                result["MapToCopy"] += removed_maps

        if "GT4PyDeadDataflowElimination" not in skip:
            eliminate_dead_dataflow_res = gtx_transformations.gt_eliminate_dead_dataflow(
                sdfg=sdfg,
                run_simplify=False,
                validate=False,
                validate_all=validate_all,
            )
            if eliminate_dead_dataflow_res != 0:
                at_least_one_xtrans_run = True
                result = result or {}
                if "GT4PyDeadDataflowElimination" not in result:
                    result["GT4PyDeadDataflowElimination"] = 0
                result["GT4PyDeadDataflowElimination"] += eliminate_dead_dataflow_res

        if "TrivialTaskletElimination" not in skip:
            eliminated_trivial_tasklets = sdfg.apply_transformations_once_everywhere(
                dace.transformation.dataflow.TrivialTaskletElimination(),
                validate=False,
                validate_all=validate_all,
            )
            if eliminated_trivial_tasklets:
                at_least_one_xtrans_run = True
                result = result or {}
                if "TrivialTaskletElimination" not in result:
                    result["TrivialTaskletElimination"] = 0
                result["TrivialTaskletElimination"] += eliminated_trivial_tasklets

        if "CopyChainRemover" not in skip:
            copy_chain_remover_result = gtx_transformations.gt_remove_copy_chain(
                sdfg=sdfg,
                validate=False,
                validate_all=validate_all,
            )
            if copy_chain_remover_result is not None:
                at_least_one_xtrans_run = True
                result = result or {}
                if "CopyChainRemover" not in result:
                    result["CopyChainRemover"] = 0
                result["CopyChainRemover"] += copy_chain_remover_result

        if "SingleStateGlobalDirectSelfCopyElimination" not in skip:
            direct_self_copy_removal_result = sdfg.apply_transformations_repeated(
                gtx_transformations.SingleStateGlobalDirectSelfCopyElimination(),
                validate=False,
                validate_all=validate_all,
            )
            if direct_self_copy_removal_result > 0:
                at_least_one_xtrans_run = True
                result = result or {}
                if "SingleStateGlobalDirectSelfCopyElimination" not in result:
                    result["SingleStateGlobalDirectSelfCopyElimination"] = 0
                result["SingleStateGlobalDirectSelfCopyElimination"] += (
                    direct_self_copy_removal_result
                )

        if "SingleStateGlobalSelfCopyElimination" not in skip:
            self_copy_removal_result = sdfg.apply_transformations_repeated(
                gtx_transformations.SingleStateGlobalSelfCopyElimination(),
                validate=False,
                validate_all=validate_all,
            )
            if self_copy_removal_result > 0:
                at_least_one_xtrans_run = True
                result = result or {}
                if "SingleStateGlobalSelfCopyElimination" not in result:
                    result["SingleStateGlobalSelfCopyElimination"] = 0
                result["SingleStateGlobalSelfCopyElimination"] += self_copy_removal_result

        if "MultiStateGlobalSelfCopyElimination" not in skip:
            distributed_self_copy_result = (
                gtx_transformations.gt_multi_state_global_self_copy_elimination(
                    sdfg, validate=validate_all
                )
            )
            if distributed_self_copy_result is not None:
                at_least_one_xtrans_run = True
                result = result or {}
                if "MultiStateGlobalSelfCopyElimination" not in result:
                    result["MultiStateGlobalSelfCopyElimination"] = set()
                result["MultiStateGlobalSelfCopyElimination"].update(distributed_self_copy_result)

    if validate:
        sdfg.validate()

    return result


def gt_inline_nested_sdfg(
    sdfg: dace.SDFG,
    permissive: bool = False,
    validate: bool = True,
    validate_all: bool = False,
) -> Optional[dict[str, int]]:
    """Perform inlining of nested SDFG into their parent SDFG.

    The function uses DaCe's `InlineSDFG` transformation, the same used in simplify.
    However, before the inline transformation is run the function will run some
    cleaning passes that allows inlining nested SDFGs.
    As a side effect, the function will split stages into more states.

    Args:
        sdfg: The SDFG that should be processed, will be modified in place and returned.
        permissive: Be less strict on the accepted SDFGs.
        validate: Perform validation after the transformation has finished.
        validate_all: Performs extensive validation.

    Note:
        - This function guarantees stable processing order, if the name of the nested
            SDFGs and the name of the state they are located in, is stable.
        - The `no_inline` attribute of the `NestedSDFG` flag only affects the inlining
            of that specific node. The clean up and inlining of nested SDFGs inside
            such an SDFG are still performed.
        - DaCe has three(!) inliner. First `InlineMultistateSDFG`, that we employ,
            secondly `InlineSDFG`, which is only capable of inlining a nested SDFG
            with a single state and `InlineSDFGs` which combines the two. However,
            `InlineSDFG` has a bug and the processing order of `InlineSDFGs` is not
            stable. Thus GT4Py has to implement its own version.
    """
    # Finding all nested SDFGs on this level.
    nested_sdfgs_to_process: list[dace_nodes.NestedSDFG] = []
    for state in sdfg.states():
        nested_sdfgs_to_process.extend(
            node for node in state.nodes() if isinstance(node, dace_nodes.NestedSDFG)
        )

    # If there are no SDFGs to process then we exit.
    if len(nested_sdfgs_to_process) == 0:
        return None

    # Now order them, such that we can process them in a stable way.
    nested_sdfgs_to_process.sort(key=lambda nsdfg: (str(nsdfg.label), str(nsdfg.sdfg.parent.label)))

    nb_preproccess_total = 0
    nb_inlines_total = 0

    # Now we start inlining all the nested SDFGs.
    #  Before a nested SDFG is inlined the function first tries to inline all
    #  SDFGs that are nested inside it, i.e. they are processed in a stable
    #  DFS order.
    for nsdfg_node in nested_sdfgs_to_process:
        nested_sdfg: dace.SDFG = nsdfg_node.sdfg
        parent_state = nested_sdfg.parent
        parent_sdfg = parent_state.sdfg
        parent_state_id = parent_state.block_id

        # Recursive processing of nested SDFGs.
        recursive_result = gt_inline_nested_sdfg(
            sdfg=nsdfg_node.sdfg,
            permissive=permissive,
            validate=False,
            validate_all=validate_all,
        )
        if recursive_result is not None:
            nb_preproccess_total += recursive_result.get("PruneSymbols|PruneConnectors", 0)
            nb_inlines_total += recursive_result.get("InlineSDFGs", 0)

        # Now perform some cleaning on the nested SDFG.
        for xform in [dace_dataflow.PruneSymbols, dace_dataflow.PruneConnectors]:
            candidate = {xform.nsdfg: nsdfg_node}
            cleaner = xform()
            cleaner.setup_match(
                sdfg=parent_sdfg,
                cfg_id=parent_state.parent_graph.cfg_id,
                state_id=parent_state_id,
                subgraph=candidate,
                expr_index=0,
                override=True,
            )
            if cleaner.can_be_applied(parent_state, 0, parent_sdfg, permissive=False):
                cleaner.apply(parent_state, parent_sdfg)
                nb_preproccess_total += 1

        # Inlining an SDFG is only possible if the nested SDFG node is at global scope.
        if parent_state.scope_dict()[nsdfg_node] is not None:
            continue

        # Check the `no_inline` flag. Note that it has to be checked here and not
        #  before to ensure that the node is recursively processed and the pruning
        #  transformations are applied.
        if nsdfg_node.no_inline:
            continue

        # Now perform the actual inlining.
        # NOTE: In [PR#2178](https://github.com/GridTools/gt4py/pull/2178) this function was
        #   modified to be more efficient. It also changed the order in which the inlining
        #   transformations of DaCe were applied. Instead of trying `InlineMultistateSDFG`
        #   it changed that such that `InlineSDFG` was used. However, this triggered
        #   [issue#2108](https://github.com/spcl/dace/issues/2108) which lead to the removals
        #   of some writes. As a temporary solution we no longer use `InlineSDFG` but only
        #   the multistate version.
        multi_state_candidate = {dace_interstate.InlineMultistateSDFG.nested_sdfg: nsdfg_node}
        multi_state_inliner = dace_interstate.InlineMultistateSDFG()
        multi_state_inliner.setup_match(
            sdfg=parent_sdfg,
            cfg_id=parent_state.parent_graph.cfg_id,
            state_id=parent_state_id,
            subgraph=multi_state_candidate,
            expr_index=0,
            override=True,
        )
        if multi_state_inliner.can_be_applied(parent_state, 0, parent_sdfg, permissive=permissive):
            multi_state_inliner.apply(parent_state, parent_sdfg)
            nb_inlines_total += 1
            if nsdfg_node.label.startswith("scan_"):
                # See `gtir_to_sdfg_scan.py::translate_scan()` for more information.
                warnings.warn(
                    f"Inlined '{nsdfg_node.label}' which might be a scan, this might leads to errors during simplification.",
                    stacklevel=0,
                )

    result: dict[str, int] = {}
    if nb_inlines_total != 0:
        result["InlineSDFGs"] = nb_inlines_total
    if nb_preproccess_total != 0:
        result["PruneSymbols|PruneConnectors"] = nb_preproccess_total

    if validate or validate_all:
        sdfg.validate()

    return result if result else None


def gt_substitute_compiletime_symbols(
    sdfg: dace.SDFG,
    repl: dict[str, Any],
    simplify: bool = False,
    skip: Optional[set[str]] = None,
    validate: bool = True,
    validate_all: bool = False,
    **kwargs: Any,
) -> None:
    """Substitutes symbols that are known at compile time with their value.

    Some symbols are known to have a constant value. This function will remove these
    symbols from the SDFG and replace them with the value.
    An example where this makes sense are strides that are known to be one.

    Args:
        sdfg: The SDFG to process.
        repl: Maps the name of the symbol to the value it should be replaced with.
        simplify: If `False` do not call `gt_simplify()` after the substitution.
        skip: List of simplify stages that should not run. Is passed to the `skip` argument
            of `gt_simplify()`.
        validate: Perform validation at the end of the function.
        validate_all: Perform validation also on intermediate steps.

    Note:
        Due to [DaCe issue 1817](https://github.com/spcl/dace/issues/1817) this function
        runs `gt_simplify()` on the SDFG. It will do so with also enable some passes
        that are by default disabled.

    Todo:
        This function needs improvement.
    """

    # NOTE: If a symbol, that should be replaced with a constant, is a data, i.e. has
    #   an entry in `sdfg.arrays` _and_ an AccessNode, then constant substitution
    #   fails. Furthermore, there is [DaCe issue 1817](https://github.com/spcl/dace/issues/1817),
    #   causing problems if there are two states in a nested SDFG. To avoid them
    #   we initially call simplify and hope for the best.
    # For testing purposes we need to be able to disable this initial simplify.
    #  This is an implementation detail that we should get rid of.
    if kwargs.get("simplify_at_entry", True):
        # NOTE: To ensure uniform behaviour and as a performance optimization,
        #   `gt_auto_optimizer()` performs the initial simplification before this
        #   function is called. If something is changed here then the change might
        #   be necessary to be applied there as well.
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            skip=skip,
            validate=False,
            validate_all=validate_all,
        )

    # We will use the `replace` function of the top SDFG, however, lower levels
    #  are handled using ConstantPropagation.
    # TODO(phimuell, iomaganaris): Revisit once the replace function in DaCe has been updated.
    sdfg.replace_dict(repl)

    # TODO(phimuell): Get rid of the `ConstantPropagation`
    const_prop = dace_passes.ConstantPropagation()
    const_prop.recursive = True
    const_prop.progress = False

    const_prop.apply_pass(
        sdfg=sdfg,
        initial_symbols=repl,
        _=None,
    )
    if simplify:
        gt_simplify(
            sdfg=sdfg,
            skip=skip,
            validate=False,
            validate_all=validate_all,
        )
    else:
        dace_sdutils.canonicalize_memlet_trees(sdfg)
    dace.sdfg.propagation.propagate_memlets_sdfg(sdfg)

    if validate:
        sdfg.validate()


def gt_reduce_distributed_buffering(
    sdfg: dace.SDFG,
    validate: bool = True,
    validate_all: bool = False,
) -> Optional[dict[dace.SDFG, dict[dace.SDFGState, set[str]]]]:
    """Removes distributed write back buffers."""
    pipeline = dace_ppl.Pipeline([DistributedBufferRelocator()])
    all_result = {}

    for rsdfg in sdfg.all_sdfgs_recursive():
        ret = pipeline.apply_pass(sdfg, {})
        if ret is not None:
            all_result[rsdfg] = ret

        if validate_all:
            rsdfg.validate()

    if len(all_result) == 0:
        return None

    if validate:
        sdfg.validate()

    return all_result


AccessLocation: TypeAlias = tuple[dace_nodes.AccessNode, dace.SDFGState]
"""Describes an access node and the state in which it is located.
"""


@dace_properties.make_properties
class DistributedBufferRelocator(dace_transformation.Pass):
    """Moves the final write back of the results to where it is needed.

    In certain cases, especially in case where we have `if` the result is computed
    in each branch and then in the join state written back. Thus there is some
    additional storage needed.
    The transformation will look for the following situation:
    - A transient data container, called `temp_storage`, is written into another
        container, called `dest_storage`, which is not transient.
    - The access node of `temp_storage` has an in degree of zero and an out degree of one.
    - The access node of `dest_storage` has an in degree of of one and an
        out degree of zero (this might be lifted).
    - `temp_storage` is not used afterwards.
    - `dest_storage` is only used to implement the buffering.

    The function will relocate the writing of `dest_storage` to where `temp_storage` is
    written, which might be multiple locations.
    It will also remove the writing back.
    It is advised that after this transformation simplify is run again.

    The relocation will not take place if it might create data race. A necessary
    but not sufficient condition for a data race is if `dest_storage` is present
    in the state where `temp_storage` is defined. In addition at least one of the
    following conditions has to be met:
    - There are accesses to `dest_storage` that are not predecessor to the node where
        the data is stored inside `temp_storage`. This check will ignore empty Memlets.
    - There is a `dest_storage` access node, that has an output degree larger
        than one.

    Note:
        - Essentially this transformation removes the double buffering of
            `dest_storage`. Because we ensure that that `dest_storage` is non
            transient this is okay, as our rule guarantees this.

    Todo:
        - Completely remove `temp_storage`, i.e. write directly into `dest_storage`.
        - Allow that `dest_storage` can also be transient.
        - Allow that `dest_storage` does not need to be a sink node, this is most
            likely most relevant if it is transient.
        - Check if `dest_storage` is used between where we want to place it and
            where it is currently used.
    """

    def modifies(self) -> dace_ppl.Modifies:
        return dace_ppl.Modifies.Memlets | dace_ppl.Modifies.AccessNodes

    def should_reapply(self, modified: dace_ppl.Modifies) -> bool:
        return modified & (dace_ppl.Modifies.Memlets | dace_ppl.Modifies.AccessNodes)

    def depends_on(self) -> set[type[dace_transformation.Pass]]:
        return {
            dace_transformation.passes.StateReachability,
            dace_transformation.passes.FindAccessStates,
        }

    def apply_pass(
        self, sdfg: dace.SDFG, pipeline_results: dict[str, Any]
    ) -> Optional[dict[dace.SDFGState, set[str]]]:
        # NOTE: We can not use `AccessSets` because this pass operates on
        #  `ControlFlowBlock`s, which might consists of multiple states. Thus we are
        #  using `FindAccessStates` which has this `SDFGState` granularity. The downside
        #  is, however, that we have to determine if the access in that state is a
        #  write or not, which means we have to find it first.
        access_states: dict[str, set[dace.SDFGState]] = pipeline_results["FindAccessStates"][
            sdfg.cfg_id
        ]

        # For speeding up the `is_accessed_downstream()` calls.
        reachable: dict[dace.SDFGState, set[dace.SDFGState]] = pipeline_results[
            "StateReachability"
        ][sdfg.cfg_id]

        result: dict[dace.SDFGState, set[str]] = collections.defaultdict(set)

        to_relocate = self._find_candidates(sdfg, reachable, access_states)
        if len(to_relocate) == 0:
            return None
        self._relocate_write_backs(sdfg, to_relocate)

        for (wb_an, wb_state), _ in to_relocate:
            result[wb_state].add(wb_an.data)

        return result

    def _relocate_write_backs(
        self,
        sdfg: dace.SDFG,
        to_relocate: list[tuple[AccessLocation, list[AccessLocation]]],
    ) -> None:
        """Perform the actual relocation."""
        for (wb_an, wb_state), def_locations in to_relocate:
            # Get the memlet that we have to replicate.
            wb_edge = next(iter(wb_state.out_edges(wb_an)))
            wb_memlet: dace.Memlet = wb_edge.data
            final_dest_name: str = wb_edge.dst.data

            for def_an, def_state in def_locations:
                # TODO(phimuell): Do not create a copy from `temp_storage` to the newly
                #   created `dest_storage`. Instead bypass `temp_storage` fully.
                def_state.add_edge(
                    def_an,
                    wb_edge.src_conn,
                    def_state.add_access(final_dest_name, copy.copy(wb_edge.dst.debuginfo)),
                    wb_edge.dst_conn,
                    copy.deepcopy(wb_memlet),
                )

            # Now remove the old node and if the old target become isolated
            #  remove that as well.
            old_dst = wb_edge.dst
            wb_state.remove_node(wb_an)
            if wb_state.degree(old_dst) == 0:
                wb_state.remove_node(old_dst)

    def _find_candidates(
        self,
        sdfg: dace.SDFG,
        reachable: dict[dace.SDFGState, set[dace.SDFGState]],
        access_states: dict[str, set[dace.SDFGState]],
    ) -> list[tuple[AccessLocation, list[AccessLocation]]]:
        """Determines all temporaries that have to be relocated.

        Returns:
            A list of tuples. The first element element of the tuple is an
            `AccessLocation` that describes where the temporary is read.
            The second element is a list of `AccessLocation`s that describes
            where the temporary is defined.
        """
        # All nodes that are used as distributed buffers.
        candidate_temp_storage: list[AccessLocation] = []

        # Which `temp_storage` access node is written back to which global memory.
        temp_storage_to_global: dict[dace_nodes.AccessNode, str] = {}

        for state in sdfg.states():
            # These are the possible targets we want to write into.
            candidate_dst_nodes: set[dace_nodes.AccessNode] = {
                node
                for node in state.sink_nodes()
                if (
                    isinstance(node, dace_nodes.AccessNode)
                    and state.in_degree(node) == 1
                    and (not node.desc(sdfg).transient)
                )
            }
            if len(candidate_dst_nodes) == 0:
                continue

            for temp_storage in state.data_nodes():
                if not temp_storage.desc(sdfg).transient:
                    continue
                if state.out_degree(temp_storage) != 1:
                    continue
                dst_candidate: dace_nodes.AccessNode = next(
                    iter(edge.dst for edge in state.out_edges(temp_storage))
                )
                if dst_candidate not in candidate_dst_nodes:
                    continue
                candidate_temp_storage.append((temp_storage, state))
                temp_storage_to_global[temp_storage] = dst_candidate.data

        if len(candidate_temp_storage) == 0:
            return []

        # Now we have to find the places where the temporary sources are defined.
        #  I.e. This is also the location where the temporary source was initialized.
        result_candidates: list[tuple[AccessLocation, list[AccessLocation]]] = []

        def find_upstream_states(dst_state: dace.SDFGState) -> set[dace.SDFGState]:
            return {
                src_state
                for src_state in sdfg.states()
                if dst_state in reachable[src_state] and dst_state is not src_state
            }

        for temp_storage in candidate_temp_storage:
            temp_storage_node, temp_storage_state = temp_storage
            def_locations: list[AccessLocation] = []
            for upstream_state in find_upstream_states(temp_storage_state):
                if self._is_written_to_in_state(
                    data=temp_storage_node.data,
                    state=upstream_state,
                    access_states=access_states,
                ):
                    # NOTE: We do not impose any restriction on `temp_storage`. Thus
                    #   It could be that we do read from it (we can never write to it)
                    #   in this state or any other state later.
                    # TODO(phimuell): Should we require that `temp_storage` is a sink
                    #   node? It might prevent or allow other optimizations.
                    new_locations = [
                        (data_node, upstream_state)
                        for data_node in upstream_state.data_nodes()
                        if data_node.data == temp_storage_node.data
                    ]
                    def_locations.extend(new_locations)
            if len(def_locations) != 0:
                result_candidates.append((temp_storage, def_locations))

        # This transformation removes `temp_storage` by writing its content directly
        #  to `dest_storage`, at the point where it is defined.
        #  For this transformation to be valid the following conditions have to be met:
        #   - Between the definition of `temp_storage` and the write back to `dest_storage`,
        #       `dest_storage` can not be accessed.
        #   - Between the definitions of `temp_storage` and the point where it is written
        #       back, `temp_storage` can only be accessed in the range that is written back.
        #   - After the write back point, `temp_storage` shall not be accessed. This
        #       restriction could be lifted.
        #
        #  To keep the implementation simple, we use the conditions:
        #   - `temp_storage` is only accessed were it is defined and at the write back
        #       point.
        #   - Between the definitions of `temp_storage` and the write back point,
        #       `dest_storage` is not used.

        result: list[tuple[AccessLocation, list[AccessLocation]]] = []

        for wb_location, def_locations in result_candidates:
            # Get the state and the location where the temporary is written back
            #  into the global data container.
            wb_node, wb_state = wb_location

            for def_node, def_state in def_locations:
                # Test if `temp_storage` is only accessed where it is defined and
                #  where it is written back.
                if gtx_transformations.utils.is_accessed_downstream(
                    start_state=def_state,
                    sdfg=sdfg,
                    reachable_states=reachable,
                    data_to_look=wb_node.data,
                    nodes_to_ignore={def_node, wb_node},
                ):
                    break

                # Check if the global data is not used between the definition of
                #  `dest_storage` and where its written back. However, we ignore
                #  the state were `temp_storage` is defined. The checks if these
                #  checks are performed by the `_check_read_write_dependency()`
                #  function.
                global_data_name = temp_storage_to_global[wb_node]
                global_nodes_in_def_state = {
                    dnode for dnode in def_state.data_nodes() if dnode.data == global_data_name
                }

                # The `is_accessed_downstream()` function has some odd behaviour
                #  regarding `states_to_ignore`. Because of the special SDFGs we have
                #  this should not be an issue.
                if gtx_transformations.utils.is_accessed_downstream(
                    start_state=def_state,
                    sdfg=sdfg,
                    reachable_states=reachable,
                    data_to_look=global_data_name,
                    nodes_to_ignore=global_nodes_in_def_state,
                    states_to_ignore={wb_state},
                ):
                    break
                if self._check_read_write_dependency(sdfg, wb_location, def_locations):
                    break
            else:
                result.append((wb_location, def_locations))

        return result

    def _is_written_to_in_state(
        self,
        data: str,
        state: dace.SDFGState,
        access_states: dict[str, set[dace.SDFGState]],
    ) -> bool:
        """This function determines if there is a write to data `data` in state `state`.

        Args:
            data: Name of the data descriptor that should be tested.
            state: The state that should be examined.
            access_states: The set of state that writes to a specific data.
        """
        assert data in access_states, f"Did not found '{data}' in 'access_states'."

        # According to `access_states` `data` is not accessed inside `state`.
        #  Therefore there is no write.
        if state not in access_states[data]:
            return False

        # There is an AccessNode for `data` inside `state`. Now we have to find the
        #  node and determine if it is a write or not.
        for dnode in state.data_nodes():
            if dnode.data != data:
                continue
            if state.in_degree(dnode) > 0:
                return True

        return False

    def _check_read_write_dependency(
        self,
        sdfg: dace.SDFG,
        write_back_location: AccessLocation,
        target_locations: list[AccessLocation],
    ) -> bool:
        """Tests if read-write conflicts would be created.

        This function ensures that the substitution of `write_back_location` into
        `target_locations` will not create a read-write conflict.
        The rules that are used for this are outlined in the class description.

        Args:
            sdfg: The SDFG on which we operate.
            write_back_location: Where currently the write back occurs.
            target_locations: List of the locations where we would like to perform
                the write back instead.

        Returns:
            If a read-write dependency is detected then the function will return
            `True` and if none was detected `False` will be returned.
        """
        for target_location in target_locations:
            if self._check_read_write_dependency_impl(sdfg, write_back_location, target_location):
                return True
        return False

    def _check_read_write_dependency_impl(
        self,
        sdfg: dace.SDFG,
        write_back_location: AccessLocation,
        target_location: AccessLocation,
    ) -> bool:
        """Tests if read-write conflict would be created for a single location.

        Args:
            sdfg: The SDFG on which we operate.
            write_back_location: Where currently the write back occurs.
            target_locations: Location where the new write back should be performed.

        Todo:
            Refine these checks later.

        Returns:
            If a read-write dependency is detected then the function will return
            `True` and if none was detected `False` will be returned.
        """
        assert write_back_location[0].data == target_location[0].data

        # Get the state and the location where the temporary is written back
        #  into the global data container. Because `write_back_node` refers to
        #  the temporary we must query the graph to find the global node.
        write_back_node, write_back_state = write_back_location
        write_back_edge = next(iter(write_back_state.out_edges(write_back_node)))
        global_data_name = write_back_edge.dst.data
        assert not sdfg.arrays[global_data_name].transient
        assert write_back_state.out_degree(write_back_node) == 1
        assert write_back_state.in_degree(write_back_node) == 0

        # Get the location and the state where the temporary is originally defined.
        def_location_of_intermediate, state_to_inspect = target_location

        # These are all access nodes that refers to the global data, that we want
        #  to move into the state `state_to_inspect`. We need them to do the
        #  second test.
        accesses_to_global_data: set[dace_nodes.AccessNode] = set()

        # In the first check we look for an access node, to the global data, that
        #  has an output degree larger than one. However, for this we ignore all
        #  empty Memlets. This is done because such Memlets are used to induce a
        #  schedule or order in the dataflow graph.
        #  As a byproduct, for the second test, we also collect all of these nodes.
        # TODO(phimuell): Refine this such that it takes the location of the data
        #   into account.
        for dnode in state_to_inspect.data_nodes():
            if dnode.data != global_data_name:
                continue
            dnode_degree = sum(
                (1 for oedge in state_to_inspect.out_edges(dnode) if not oedge.data.is_empty())
            )
            if dnode_degree > 1:
                return True
            # TODO(phimuell): Maybe AccessNodes with zero input degree should be ignored.
            accesses_to_global_data.add(dnode)

        # There is no reference to the global data, so no need to do more tests.
        if len(accesses_to_global_data) == 0:
            return False

        # For the second test we will explore the dataflow graph, in reverse order,
        #  starting from the definition of the temporary node. If we find an access
        #  to the global data we remove it from the `accesses_to_global_data` list.
        #  If the list has not become empty, then we know that there is some sind
        #  branch (or concurrent dataflow) in this state that accesses the global
        #  data and we will have read-write conflicts.
        #  It is however, important to realize that passing this check does not
        #  imply that there are no read-write. We assume here that all accesses to
        #  the global data that was made before the write back were constructed in
        #  a correct way.
        to_process: list[dace_nodes.Node] = [def_location_of_intermediate]
        seen: set[dace_nodes.Node] = set()
        while len(to_process) != 0:
            node = to_process.pop()
            seen.add(node)

            if isinstance(node, dace_nodes.AccessNode):
                if node.data == global_data_name:
                    accesses_to_global_data.discard(node)
                    if len(accesses_to_global_data) == 0:
                        return False

            # Note that we only explore the ingoing edges, thus we will not necessarily
            #  explore the whole graph. However, this is fine, because we will see the
            #  relevant parts. To see that assume that we would also have to check the
            #  outgoing edges, this would mean that there was some branching point,
            #  which is a serialization point, so the dataflow would have been invalid
            #  before.
            to_process.extend(
                iedge.src for iedge in state_to_inspect.in_edges(node) if iedge.src not in seen
            )

        assert len(accesses_to_global_data) > 0
        return True


@dace_properties.make_properties
class GT4PyMoveTaskletIntoMap(dace_transformation.SingleStateTransformation):
    """Moves a Tasklet, with no input into a map.

    Tasklets without inputs, are mostly used to generate constants.
    However, if they are outside a Map, then this constant value is an
    argument to the kernel, and can not be used by the compiler.

    This transformation moves such Tasklets into a Map scope.
    """

    tasklet = dace_transformation.PatternNode(dace_nodes.Tasklet)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.tasklet, cls.access_node, cls.map_entry)]

    def can_be_applied(
        self,
        graph: dace.SDFGState | dace.SDFG,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        tasklet: dace_nodes.Tasklet = self.tasklet
        access_node: dace_nodes.AccessNode = self.access_node
        access_desc: dace_data.Data = access_node.desc(sdfg)
        map_entry: dace_nodes.MapEntry = self.map_entry

        if graph.in_degree(tasklet) != 0:
            return False
        if graph.out_degree(tasklet) != 1:
            return False
        if tasklet.has_side_effects(sdfg):
            return False
        if tasklet.code_init.as_string:
            return False
        if tasklet.code_exit.as_string:
            return False
        if tasklet.code_global.as_string:
            return False
        if tasklet.state_fields:
            return False
        if not isinstance(access_desc, dace_data.Scalar):
            return False
        if not access_desc.transient:
            return False
        if not any(
            edge.dst_conn and edge.dst_conn.startswith("IN_")
            for edge in graph.out_edges(access_node)
            if edge.dst is map_entry
        ):
            return False
        # NOTE: We allow that the access node is used in multiple places.

        return True

    def apply(
        self,
        graph: dace.SDFGState | dace.SDFG,
        sdfg: dace.SDFG,
    ) -> None:
        tasklet: dace_nodes.Tasklet = self.tasklet
        access_node: dace_nodes.AccessNode = self.access_node
        access_desc: dace_data.Scalar = access_node.desc(sdfg)
        map_entry: dace_nodes.MapEntry = self.map_entry

        # Find _a_ connection that leads from the access node to the map.
        edge_to_map = next(
            iter(
                edge
                for edge in graph.out_edges(access_node)
                if edge.dst is map_entry and edge.dst_conn.startswith("IN_")
            )
        )
        connector_name: str = edge_to_map.dst_conn[3:]

        # This is the tasklet that we will put inside the map, note we have to do it
        #  this way to avoid some name clash stuff.
        inner_tasklet: dace_nodes.Tasklet = graph.add_tasklet(
            name=f"{tasklet.label}__clone_{str(uuid.uuid1()).replace('-', '_')}",
            outputs=tasklet.out_connectors.keys(),
            inputs=set(),
            code=tasklet.code,
            language=tasklet.language,
            debuginfo=tasklet.debuginfo,
        )
        inner_desc: dace_data.Scalar = access_desc.clone()
        inner_data_name: str = sdfg.add_datadesc(access_node.data, inner_desc, find_new_name=True)
        inner_an: dace_nodes.AccessNode = graph.add_access(
            inner_data_name, copy.copy(access_node.debuginfo)
        )

        # Connect the tasklet with the map entry and the access node.
        graph.add_nedge(map_entry, inner_tasklet, dace.Memlet())
        graph.add_edge(
            inner_tasklet,
            next(iter(inner_tasklet.out_connectors.keys())),
            inner_an,
            None,
            dace.Memlet(f"{inner_data_name}[0]"),
        )

        # Now we will reroute the edges went through the inner map, through the
        #  inner access node instead.
        for old_inner_edge in list(
            graph.out_edges_by_connector(map_entry, "OUT_" + connector_name)
        ):
            # We now modify the downstream data. This is because we no longer refer
            #  to the data outside but the one inside.
            self._modify_downstream_memlets(
                state=graph,
                edge=old_inner_edge,
                old_data=access_node.data,
                new_data=inner_data_name,
            )

            # After we have changed the properties of the MemletTree of `edge`
            #  we will now reroute it, such that the inner access node is used.
            graph.add_edge(
                inner_an,
                None,
                old_inner_edge.dst,
                old_inner_edge.dst_conn,
                old_inner_edge.data,
            )
            graph.remove_edge(old_inner_edge)
            map_entry.remove_in_connector("IN_" + connector_name)
            map_entry.remove_out_connector("OUT_" + connector_name)

        # Now we can remove the map connection between the outer/old access
        #  node and the map.
        graph.remove_edge(edge_to_map)

        # The data is no longer referenced in this state, so we can potentially
        #  remove
        if graph.out_degree(access_node) == 0:
            # TODO(phimuell): Use the pipeline to run `StateReachability` once.
            if not gtx_transformations.utils.is_accessed_downstream(
                start_state=graph,
                sdfg=sdfg,
                reachable_states=None,
                data_to_look=access_node.data,
                nodes_to_ignore={access_node},
            ):
                graph.remove_nodes_from([tasklet, access_node])
                # Needed if data is accessed in a parallel branch.
                try:
                    sdfg.remove_data(access_node.data, validate=True)
                except ValueError as e:
                    if not str(e).startswith(f"Cannot remove data descriptor {access_node.data}:"):
                        raise

    def _modify_downstream_memlets(
        self,
        state: dace.SDFGState,
        edge: dace.sdfg.graph.MultiConnectorEdge,
        old_data: str,
        new_data: str,
    ) -> None:
        """Replaces the data along on the tree defined by `edge`.

        The function will traverse the MemletTree defined by `edge`.
        Any Memlet that refers to `old_data` will be replaced with
        `new_data`.

        Args:
            state: The sate in which we operate.
            edge: The edge defining the MemletTree.
            old_data: The name of the data that should be replaced.
            new_data: The name of the new data the Memlet should refer to.
        """
        mtree: dace.memlet.MemletTree = state.memlet_tree(edge)
        for tedge in mtree.traverse_children(True):
            # Because we only change the name of the data, we do not change the
            #  direction of the Memlet, so `{src, dst}_subset` will remain the same.
            if tedge.edge.data.data == old_data:
                tedge.edge.data.data = new_data


@dace_properties.make_properties
class GT4PyMapBufferElimination(dace_transformation.SingleStateTransformation):
    """Allows to remove unneeded buffering at map output.

    The transformation matches the case `MapExit -> (T) -> (G)`, where `T` is an
    AccessNode referring to a transient and `G` an AccessNode that refers to non
    transient memory.
    If the following conditions are met then `T` is removed.
    - `T` is not used to filter computations, i.e. what is written into `G`
        is covered by what is written into `T`.
    - `T` is not used anywhere else.
    - `G` is not also an input to the map, except there is only a pointwise
        dependency in `G`, see the note below.
    - Everything needs to be at top scope.

    Notes:
        - Rule 3 of ADR18 should guarantee that any valid GT4Py program meets the
            point wise dependency in `G`, for that reason it is possible to disable
            this test by specifying `assume_pointwise`.

    Todo:
        - Implement a real pointwise test.
        - Run this inside a pipeline.
    """

    map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    tmp_ac = dace_transformation.transformation.PatternNode(dace_nodes.AccessNode)
    glob_ac = dace_transformation.PatternNode(dace_nodes.AccessNode)

    assume_pointwise = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Dimensions that should become the leading dimension.",
    )

    def __init__(
        self,
        assume_pointwise: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if assume_pointwise is not None:
            self.assume_pointwise = assume_pointwise

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.map_exit, cls.tmp_ac, cls.glob_ac)]

    def depends_on(self) -> set[type[dace_transformation.Pass]]:
        return {dace_transformation.passes.ConsolidateEdges}

    def can_be_applied(
        self,
        graph: dace.SDFGState | dace.SDFG,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        tmp_ac: dace_nodes.AccessNode = self.tmp_ac
        glob_ac: dace_nodes.AccessNode = self.glob_ac
        tmp_desc: dace_data.Data = tmp_ac.desc(sdfg)
        glob_desc: dace_data.Data = glob_ac.desc(sdfg)

        if not tmp_desc.transient:
            return False
        if glob_desc.transient:
            return False
        if graph.in_degree(tmp_ac) != 1:
            return False
        if any(gtx_transformations.utils.is_view(ac, sdfg) for ac in [tmp_ac, glob_ac]):
            return False
        if len(glob_desc.shape) != len(tmp_desc.shape):
            return False

        # Test if we are on the top scope (it is likely).
        if graph.scope_dict()[glob_ac] is not None:
            return False

        # Now perform if we are point wise
        if not self._perform_pointwise_test(graph, sdfg):
            return False

        # Test if `tmp` is only anywhere else, this is important for removing it.
        if graph.out_degree(tmp_ac) != 1:
            return False
        # TODO(phimuell): Use the pipeline system to run the `StateReachability` pass
        #  only once. Taking care of DaCe issue 1911.
        if gtx_transformations.utils.is_accessed_downstream(
            start_state=graph,
            sdfg=sdfg,
            reachable_states=None,
            data_to_look=tmp_ac.data,
            nodes_to_ignore={tmp_ac},
        ):
            return False

        # Now we ensure that `tmp` is not used to filter out some computations.
        map_to_tmp_edge = next(edge for edge in graph.in_edges(tmp_ac))
        tmp_to_glob_edge = next(edge for edge in graph.out_edges(tmp_ac))

        tmp_in_subset = map_to_tmp_edge.data.get_dst_subset(map_to_tmp_edge, graph)
        tmp_out_subset = tmp_to_glob_edge.data.get_src_subset(tmp_to_glob_edge, graph)
        glob_in_subset = tmp_to_glob_edge.data.get_dst_subset(tmp_to_glob_edge, graph)
        if tmp_in_subset is None:
            tmp_in_subset = dace_subsets.Range.from_array(tmp_desc)
        if tmp_out_subset is None:
            tmp_out_subset = dace_subsets.Range.from_array(tmp_desc)
        if glob_in_subset is None:
            return False

        # TODO(phimuell): Do we need simplify in the check.
        # TODO(phimuell): Restrict this to having the same size.
        if tmp_out_subset != tmp_in_subset:
            return False
        return True

    def _perform_pointwise_test(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> bool:
        """Test if `G` is only point wise accessed.

        This function will also consider the `assume_pointwise` property.
        """
        map_exit: dace_nodes.MapExit = self.map_exit
        map_entry: dace_nodes.MapEntry = state.entry_node(map_exit)
        glob_ac: dace_nodes.AccessNode = self.glob_ac
        glob_data: str = glob_ac.data

        # First we check if `G` is also an input to this map.
        conflicting_inputs: set[dace_nodes.AccessNode] = set()
        for in_edge in state.in_edges(map_entry):
            if not isinstance(in_edge.src, dace_nodes.AccessNode):
                continue

            # Find the source of this data, if it is a view we trace it to
            #  its origin.
            src_node: dace_nodes.AccessNode = gtx_transformations.utils.track_view(
                in_edge.src, state, sdfg
            )

            # Test if there is a conflict; We do not store the source but the
            #  actual node that is adjacent.
            if src_node.data == glob_data:
                conflicting_inputs.add(in_edge.src)

        # If there are no conflicting inputs, then we are point wise.
        #  This is an implementation detail that make life simpler.
        if len(conflicting_inputs) == 0:
            return True

        # If we can assume pointwise computations, then we do not have to do
        #  anything.
        if self.assume_pointwise:
            return True

        # Currently the only test that we do is, if we have a view, then we
        #  are not point wise.
        # TODO(phimuell): Improve/implement this.
        return any(gtx_transformations.utils.is_view(node, sdfg) for node in conflicting_inputs)

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        # Removal
        # Propagation ofthe shift.
        map_exit: dace_nodes.MapExit = self.map_exit
        tmp_ac: dace_nodes.AccessNode = self.tmp_ac
        tmp_desc: dace_data.Data = tmp_ac.desc(sdfg)
        tmp_data = tmp_ac.data
        glob_ac: dace_nodes.AccessNode = self.glob_ac
        glob_data = glob_ac.data

        map_to_tmp_edge = next(edge for edge in graph.in_edges(tmp_ac))
        tmp_to_glob_edge = next(edge for edge in graph.out_edges(tmp_ac))

        glob_in_subset = tmp_to_glob_edge.data.get_dst_subset(tmp_to_glob_edge, graph)
        tmp_out_subset = tmp_to_glob_edge.data.get_src_subset(tmp_to_glob_edge, graph)
        if tmp_out_subset is None:
            tmp_out_subset = dace_subsets.Range.from_array(tmp_desc)
        assert glob_in_subset is not None

        # Recursively visit the nested SDFGs for mapping of strides from inner to outer array
        gtx_transformations.gt_map_strides_to_src_nested_sdfg(sdfg, graph, map_to_tmp_edge, glob_ac)

        # We now remove the `tmp` node, and create a new connection between
        #  the global node and the map exit.
        new_map_to_glob_edge = graph.add_edge(
            map_exit,
            map_to_tmp_edge.src_conn,
            glob_ac,
            tmp_to_glob_edge.dst_conn,
            dace.Memlet(
                data=glob_ac.data,
                subset=copy.deepcopy(glob_in_subset),
            ),
        )
        graph.remove_edge(map_to_tmp_edge)
        graph.remove_edge(tmp_to_glob_edge)
        graph.remove_node(tmp_ac)

        # We can not unconditionally remove the data `tmp` refers to, because
        #  it could be that in a parallel branch the `tmp` is also defined.
        try:
            sdfg.remove_data(tmp_ac.data, validate=True)
        except ValueError as e:
            if not str(e).startswith(f"Cannot remove data descriptor {tmp_ac.data}:"):
                raise

        # Now we must modify the memlets inside the map scope, because
        #  they now write into `G` instead of `tmp`, which has a different
        #  offset.
        # NOTE: Assumes that `tmp_out_subset` and `tmp_in_subset` are the same.
        correcting_offset = glob_in_subset.offset_new(tmp_out_subset, negative=True)
        mtree = graph.memlet_tree(new_map_to_glob_edge)
        for tree in mtree.traverse_children(include_self=False):
            curr_edge = tree.edge
            curr_dst_subset = curr_edge.data.get_dst_subset(curr_edge, graph)
            if curr_edge.data.data == tmp_data:
                curr_edge.data.data = glob_data
            if curr_dst_subset is not None:
                curr_dst_subset.offset(correcting_offset, negative=False)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import ClassVar, TypeVar

from gt4py import eve
from gt4py.next import utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import trace_shifts
from gt4py.next.iterator.transforms.inline_lambdas import inline_lambda
from gt4py.next.iterator.transforms.inline_lifts import InlineLifts


def is_center_derefed_only(node: itir.Node) -> bool:
    return hasattr(node.annex, "recorded_shifts") and node.annex.recorded_shifts in [set(), {()}]


T = TypeVar("T", bound=itir.Node)


@dataclasses.dataclass
class InlineCenterDerefLiftVars(eve.NodeTranslator):
    """
    Inline all variables which are derefed in the center only (i.e. unshifted).

    Consider the following example where `var` is never shifted:

    `let(var, (↑stencil)(it))(·var + ·var)`

    Directly inlining `var` would increase the size of the tree and duplicate the calculation.
    Instead this pass, first takes the iterator `(↑stencil)(it)` and transforms it into a
    0-ary function that evaluates to the value at the current location.

    `λ() → ·(↑stencil)(it)`

    Then all previous occurences of `var` are replaced by this function.

    `let(_icdlv_1, λ() → ·(↑stencil)(it))(·(↑(λ() → _icdlv_1()) + ·(↑(λ() → _icdlv_1()))`

    The lift inliner can then later easily transform this into a nice expression:

     `let(_icdlv_1, λ() → stencil(it))(_icdlv_1() + _icdlv_1())`

    Finally, recomputation is avoided by using the common subexpression elimination and lamba
    inlining (can be configured opcount preserving). Both is up to the caller to do later.

    `λ(_cs_1) → _cs_1 + _cs_1)(stencil(it))`

    Note: This pass uses and preserves the `domain` and `recorded_shifts` annex.
    """

    PRESERVED_ANNEX_ATTRS: ClassVar[tuple[str, ...]] = ("domain", "recorded_shifts")

    uids: utils.IDGeneratorPool

    @classmethod
    def apply(cls, node: T, *, is_stencil=False, uids: utils.IDGeneratorPool) -> T:
        if is_stencil:
            assert isinstance(node, itir.Expr)
            trace_shifts.trace_stencil(node, save_to_annex=True)
        return cls(uids=uids).visit(node)

    def visit_FunCall(self, node: itir.FunCall, **kwargs):
        # TODO(tehrengruber): move the analysis out of this pass and just make it a requirement
        #  such that we don't need to run in multiple times if multiple passes use it.
        if cpm.is_call_to(node.fun, "as_fieldop"):
            assert len(node.fun.args) in [1, 2]
            trace_shifts.trace_stencil(
                node.fun.args[0], num_args=len(node.args), save_to_annex=True
            )

        node = self.generic_visit(node)

        if cpm.is_let(node):
            assert isinstance(node.fun, itir.Lambda)  # to make mypy happy
            eligible_params = [False] * len(node.fun.params)
            new_args = []
            # values are 0-ary lambda functions that evaluate to the derefed argument. We don't put
            # the values themselves here as they might be inside of an if to protected from an oob
            # access
            evaluators: dict[str, itir.Expr] = {}

            for i, (param, arg) in enumerate(zip(node.fun.params, node.args)):
                if cpm.is_applied_lift(arg) and is_center_derefed_only(param):
                    eligible_params[i] = True
                    bound_arg_evaluator = next(self.uids["_icdlv"])
                    capture_lift = im.promote_to_const_iterator(im.call(bound_arg_evaluator)())
                    trace_shifts.copy_recorded_shifts(from_=param, to=capture_lift)
                    new_args.append(capture_lift)
                    # since we deref an applied lift here we can (but don't need to) immediately
                    # inline
                    evaluators[bound_arg_evaluator] = im.lambda_()(
                        InlineLifts(flags=InlineLifts.Flag.INLINE_DEREF_LIFT).visit(
                            im.deref(arg), recurse=False
                        )
                    )
                else:
                    new_args.append(arg)

            if any(eligible_params):
                new_node = inline_lambda(
                    im.call(node.fun)(*new_args), eligible_params=eligible_params
                )
                # TODO(tehrengruber): propagate let outwards
                return im.let(*evaluators.items())(new_node)

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from typing import Any, Generic, TypeAlias, TypeVar, Union

from gt4py import eve
from gt4py.eve import (
    Coerced,
    Node,
    SourceLocation,
    SymbolName,
    SymbolRef,
    concepts as eve_concepts,
    datamodels,
    utils as eve_utils,
)
from gt4py.eve.traits import SymbolTableTrait
from gt4py.eve.type_definitions import StrEnum
from gt4py.next.ffront import dialect_ast_enums, type_specifications as ts_ffront
from gt4py.next.type_system import type_specifications as ts
from gt4py.next.utils import RecursionGuard


#: Generate a custom pickler which ignores "location" attribute for `eve.Node`s.
nonlocated_node_pickler = eve_concepts.skipping_fields_node_pickler("location")

#: Generate an unique fingerprint for an `eve.Node` that is location agnostic.
nonlocated_fingerprint: eve_concepts.NodeFingerprinter = functools.partial(
    eve_utils.content_hash, pickler=nonlocated_node_pickler
)


class LocatedNode(Node):
    location: SourceLocation = eve.field(repr=False, compare=False)

    def fingerprint(self) -> str:
        """
        Generates a unique hash string for this node that is location agnostic.
        """
        return nonlocated_fingerprint(self)

    def __str__(self) -> str:
        from gt4py.next.ffront.foast_pretty_printer import pretty_format

        try:
            with RecursionGuard(self):
                return pretty_format(self)
        except RecursionGuard.RecursionDetected:
            # If `pretty_format` itself calls `__str__`, i.e. when printing
            # an error that happend during formatting, just return the regular
            # string representation, consequently avoiding an infinite recursion.
            # Note that avoiding circular calls to `__str__` is not possible
            # as the pretty printer depends on eve utilities that use the string
            # representation of a node in the error handling.
            return super().__str__()


SymbolT = TypeVar("SymbolT", bound=ts.TypeSpec)


# TODO(egparedes): this should be an actual generic datamodel but it is not fully working
#   due to nested specialization with bound typevars, so disabling specialization for now
#       class Symbol(eve.GenericNode, LocatedNode, Generic[SymbolT]):
#
class Symbol(LocatedNode, Generic[SymbolT]):
    id: Coerced[SymbolName]
    type: Union[SymbolT, ts.DeferredType]  # A003
    namespace: dialect_ast_enums.Namespace = dialect_ast_enums.Namespace(
        dialect_ast_enums.Namespace.LOCAL
    )


DataTypeT = TypeVar("DataTypeT", bound=ts.DataType)
DataSymbol = Symbol[DataTypeT]

FieldTypeT = TypeVar("FieldTypeT", bound=ts.FieldType)
FieldSymbol = DataSymbol[FieldTypeT]

ScalarTypeT = TypeVar("ScalarTypeT", bound=ts.ScalarType)
ScalarSymbol = DataSymbol[ScalarTypeT]

TupleTypeT = TypeVar("TupleTypeT", bound=ts.TupleType)
TupleSymbol = DataSymbol[TupleTypeT]

NamedCollectionTypeT = TypeVar("NamedCollectionTypeT", bound=ts.NamedCollectionType)
NamedCollectionSymbol = DataSymbol[NamedCollectionTypeT]

DimensionTypeT = TypeVar("DimensionTypeT", bound=ts.DimensionType)
DimensionSymbol = DataSymbol[DimensionTypeT]


class Expr(LocatedNode):
    type: ts.TypeSpec = ts.DeferredType(constraint=None)  # A003


class Name(Expr):
    id: Coerced[SymbolRef]


class Constant(Expr):
    value: Any  # TODO: be more specific


class Subscript(Expr):
    value: Expr
    index: Expr


class Attribute(Expr):
    value: Expr
    attr: str


class TupleExpr(Expr):
    elts: list[Expr]


class UnaryOp(Expr):
    op: dialect_ast_enums.UnaryOperator
    operand: Expr


class BinOp(Expr):
    op: dialect_ast_enums.BinaryOperator
    left: Expr
    right: Expr


class CompareOperator(StrEnum):
    EQ = "eq"
    NOTEQ = "not_eq"
    LT = "less"
    LTE = "less_equal"
    GT = "greater"
    GTE = "greater_equal"

    def __str__(self) -> str:
        if self is self.EQ:
            return "=="
        elif self is self.NOTEQ:
            return "!="
        elif self is self.LT:
            return "<"
        elif self is self.LTE:
            return "<="
        elif self is self.GT:
            return ">"
        elif self is self.GTE:
            return ">="
        return "Unknown CompareOperator"


class Compare(Expr):
    op: CompareOperator
    left: Expr
    right: Expr


class TernaryExpr(Expr):
    condition: Expr
    true_expr: Expr
    false_expr: Expr


class Call(Expr):
    func: Expr
    args: list[Expr]
    kwargs: dict[str, Expr]


class Stmt(LocatedNode): ...


class Starred(Expr):
    id: Union[FieldSymbol, TupleSymbol, ScalarSymbol]


class Assign(Stmt):
    target: Union[FieldSymbol, TupleSymbol, ScalarSymbol]
    value: Expr


class TupleTargetAssign(Stmt):
    targets: list[ScalarSymbol | FieldSymbol | TupleSymbol | NamedCollectionSymbol | Starred]
    value: Expr


class Return(Stmt):
    value: Expr


class BlockStmt(Stmt, SymbolTableTrait):
    stmts: list[Stmt]


class IfStmt(Stmt):
    condition: Expr
    true_branch: BlockStmt
    false_branch: BlockStmt

    @datamodels.root_validator
    @classmethod
    def _collect_common_symbols(cls: type[IfStmt], instance: IfStmt) -> None:
        common_symbol_names = sorted(  # sort is required to get stable results across runs
            instance.true_branch.annex.symtable.keys() & instance.false_branch.annex.symtable.keys()
        )
        instance.annex.propagated_symbols = {
            sym_name: Symbol(
                id=sym_name, type=ts.DeferredType(constraint=None), location=instance.location
            )
            for sym_name in common_symbol_names
        }


class FunctionDefinition(LocatedNode, SymbolTableTrait):
    id: Coerced[SymbolName]
    params: list[DataSymbol]
    body: BlockStmt
    closure_vars: list[Symbol]
    type: Union[ts.FunctionType, ts.DeferredType] = ts.DeferredType(constraint=ts.FunctionType)


class FieldOperator(LocatedNode, SymbolTableTrait):
    id: Coerced[SymbolName]
    definition: FunctionDefinition
    type: Union[ts_ffront.FieldOperatorType, ts.DeferredType] = ts.DeferredType(
        constraint=ts_ffront.FieldOperatorType
    )


class ScanOperator(LocatedNode, SymbolTableTrait):
    id: Coerced[SymbolName]
    axis: Constant
    forward: Constant
    init: Constant
    definition: FunctionDefinition  # scan pass
    type: Union[ts_ffront.ScanOperatorType, ts.DeferredType] = ts.DeferredType(
        constraint=ts_ffront.ScanOperatorType
    )


OperatorNode: TypeAlias = FieldOperator | ScanOperator



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools
from typing import Any

import factory
import numpy as np

import gt4py._core.definitions as core_defs
import gt4py.next.custom_layout_allocators as next_allocators
from gt4py._core import filecache
from gt4py.next import backend, common, config, field_utils
from gt4py.next.embedded import nd_array_field
from gt4py.next.instrumentation import metrics
from gt4py.next.otf import recipes, stages, workflow
from gt4py.next.otf.binding import nanobind
from gt4py.next.otf.compilation import compiler
from gt4py.next.otf.compilation.build_systems import compiledb
from gt4py.next.program_processors.codegens.gtfn import gtfn_module


def convert_arg(arg: Any) -> Any:
    # Note: this function is on the hot path and needs to have minimal overhead.
    if (origin := getattr(arg, "__gt_origin__", None)) is not None:
        # `Field` is the most likely case, we use `__gt_origin__` as the property is needed anyway
        # and (currently) uniquely identifies a `NDArrayField` (which is the only supported `Field`)
        assert isinstance(arg, nd_array_field.NdArrayField)
        return arg.ndarray, origin
    if isinstance(arg, tuple):
        return tuple(convert_arg(a) for a in arg)
    if isinstance(arg, np.bool_):
        # nanobind does not support implicit conversion of `np.bool` to `bool`
        return bool(arg)
    # TODO(havogt): if this function still appears in profiles,
    # we should avoid going through the previous isinstance checks for detecting a scalar.
    # E.g. functools.cache on the arg type, returning a function that does the conversion
    return arg


def convert_args(
    inp: stages.ExecutableProgram, device: core_defs.DeviceType = core_defs.DeviceType.CPU
) -> stages.ExecutableProgram:
    def decorated_program(
        *args: Any,
        offset_provider: dict[str, common.Connectivity | common.Dimension],
        out: Any = None,
    ) -> None:
        # Note: this function is on the hot path and needs to have minimal overhead.
        if out is not None:
            args = (*args, out)
        converted_args = (convert_arg(arg) for arg in args)
        conn_args = extract_connectivity_args(offset_provider, device)

        opt_kwargs: dict[str, Any] = {}
        if collect_metrics := metrics.is_level_enabled(metrics.PERFORMANCE):
            # If we are collecting metrics, we need to add the `exec_info` argument
            # to the `inp` call, which will be used to collect performance metrics.
            exec_info: dict[str, float] = {}
            opt_kwargs["exec_info"] = exec_info

        # generate implicit domain size arguments only if necessary, using `iter_size_args()`
        inp(
            *converted_args,
            *conn_args,
            **opt_kwargs,
        )

        if collect_metrics:
            metrics.add_sample_to_current_source(
                metrics.COMPUTE_METRIC, exec_info["run_cpp_duration"]
            )

    return decorated_program


def extract_connectivity_args(
    offset_provider: dict[str, common.Connectivity | common.Dimension], device: core_defs.DeviceType
) -> list[tuple[core_defs.NDArrayObject, tuple[int, ...]]]:
    # Note: this function is on the hot path and needs to have minimal overhead.
    zero_origin = (0, 0)
    assert all(
        hasattr(conn, "ndarray") or isinstance(conn, common.Dimension)
        for conn in offset_provider.values()
    )
    # Note: the order here needs to agree with the order of the generated bindings.
    # This is currently true only because when hashing offset provider dicts,
    # the keys' order is taken into account. Any modification to the hashing
    # of offset providers may break this assumption here.
    args: list[tuple[core_defs.NDArrayObject, tuple[int, ...]]] = [
        (ndarray, zero_origin)
        for conn in offset_provider.values()
        if (ndarray := getattr(conn, "ndarray", None)) is not None
    ]
    assert all(
        common.is_neighbor_table(conn) and field_utils.verify_device_field_type(conn, device)
        for conn in offset_provider.values()
        if hasattr(conn, "ndarray")
    )

    return args


class GTFNCompileWorkflowFactory(factory.Factory):
    class Meta:
        model = recipes.OTFCompileWorkflow

    class Params:
        device_type: core_defs.DeviceType = core_defs.DeviceType.CPU
        cmake_build_type: config.CMakeBuildType = factory.LazyFunction(  # type: ignore[assignment] # factory-boy typing not precise enough
            lambda: config.CMAKE_BUILD_TYPE
        )
        builder_factory: compiler.BuildSystemProjectGenerator = factory.LazyAttribute(  # type: ignore[assignment] # factory-boy typing not precise enough
            lambda o: compiledb.CompiledbFactory(cmake_build_type=o.cmake_build_type)
        )

        cached_translation = factory.Trait(
            translation=factory.LazyAttribute(
                lambda o: workflow.CachedStep(
                    o.bare_translation,
                    hash_function=stages.fingerprint_compilable_program,
                    cache=filecache.FileCache(str(config.BUILD_CACHE_DIR / "gtfn_cache")),
                )
            ),
        )

        bare_translation = factory.SubFactory(
            gtfn_module.GTFNTranslationStepFactory,
            device_type=factory.SelfAttribute("..device_type"),
        )

    translation = factory.LazyAttribute(lambda o: o.bare_translation)

    bindings: workflow.Workflow[stages.ProgramSource, stages.CompilableProject] = (
        nanobind.bind_source
    )
    compilation = factory.SubFactory(
        compiler.CompilerFactory,
        cache_lifetime=factory.LazyFunction(lambda: config.BUILD_CACHE_LIFETIME),
        builder_factory=factory.SelfAttribute("..builder_factory"),
    )
    decoration = factory.LazyAttribute(
        lambda o: functools.partial(convert_args, device=o.device_type)
    )


class GTFNBackendFactory(factory.Factory):
    class Meta:
        model = backend.Backend

    class Params:
        name_device = "cpu"
        name_cached = ""
        name_temps = ""
        name_postfix = ""
        gpu = factory.Trait(
            allocator=next_allocators.StandardGPUFieldBufferAllocator(),
            device_type=core_defs.CUPY_DEVICE_TYPE or core_defs.DeviceType.CUDA,
            name_device="gpu",
        )
        cached = factory.Trait(
            executor=factory.LazyAttribute(
                lambda o: workflow.CachedStep(o.otf_workflow, hash_function=o.hash_function)
            ),
            name_cached="_cached",
        )
        device_type = core_defs.DeviceType.CPU
        hash_function = stages.compilation_hash
        otf_workflow = factory.SubFactory(
            GTFNCompileWorkflowFactory, device_type=factory.SelfAttribute("..device_type")
        )

    name = factory.LazyAttribute(
        lambda o: f"run_gtfn_{o.name_device}{o.name_temps}{o.name_cached}{o.name_postfix}"
    )

    executor = factory.LazyAttribute(lambda o: o.otf_workflow)
    allocator = next_allocators.StandardCPUFieldBufferAllocator()
    transforms = backend.DEFAULT_TRANSFORMS


run_gtfn = GTFNBackendFactory()

run_gtfn_imperative = GTFNBackendFactory(
    name_postfix="_imperative", otf_workflow__translation__use_imperative_backend=True
)

run_gtfn_cached = GTFNBackendFactory(cached=True, otf_workflow__cached_translation=True)

run_gtfn_gpu = GTFNBackendFactory(gpu=True)

run_gtfn_gpu_cached = GTFNBackendFactory(
    gpu=True, cached=True, otf_workflow__cached_translation=True
)

run_gtfn_no_transforms = GTFNBackendFactory(
    otf_workflow__bare_translation__enable_itir_transforms=False
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from typing import ClassVar

import gt4py.eve as eve
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.transforms.symbol_ref_utils import CountSymbolRefs


class MergeLet(eve.PreserveLocationVisitor, eve.NodeTranslator):
    """
    Merge let-like statements.

    For example transforms::

        (λ(a) → (λ(b) → a+b)(arg1))(arg2)

    into::

        (λ(a, b) → a+b)(arg1, arg2)

    This can significantly reduce the depth of the tree and its readability.
    """

    PRESERVED_ANNEX_ATTRS: ClassVar[tuple[str, ...]] = ("domain",)

    def visit_FunCall(self, node: itir.FunCall):
        node = self.generic_visit(node)
        if (
            isinstance(node.fun, itir.Lambda)
            and isinstance(node.fun.expr, itir.FunCall)
            and isinstance(node.fun.expr.fun, itir.Lambda)
        ):
            outer_lambda = node.fun
            outer_lambda_args = node.args
            inner_lambda = node.fun.expr.fun
            inner_lambda_args = node.fun.expr.args
            # skip if we have a collision
            if set(outer_lambda.params) & set(inner_lambda.params):
                return node
            # check if the arguments to the outer lambda call use a symbol of the inner lambda
            # e.g. (λ(a) → (λ(b) → b)(a))(b)
            ref_counts_outer = CountSymbolRefs.apply(
                outer_lambda_args, [param.id for param in inner_lambda.params]
            )
            if any(ref_count != 0 for ref_count in ref_counts_outer.values()):
                return node
            # check if the argument to the inner lambda call depend on an argument to the outer lambda
            ref_counts = CountSymbolRefs.apply(
                inner_lambda_args, [param.id for param in outer_lambda.params]
            )
            if any(ref_count != 0 for ref_count in ref_counts.values()):
                return node
            return itir.FunCall(
                fun=itir.Lambda(
                    params=outer_lambda.params + inner_lambda.params, expr=inner_lambda.expr
                ),
                args=outer_lambda_args + inner_lambda_args,
            )
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import multiprocessing
import os
from typing import Any

import gridtools_cpp

from gt4py.cartesian.utils.compiler import cxx_compiler_defaults, gpu_configuration


GT4PY_INSTALLATION_PATH: str = os.path.dirname(os.path.abspath(__file__))

GT_INCLUDE_PATH: str = os.path.abspath(gridtools_cpp.get_include_dir())

GT_CPP_TEMPLATE_DEPTH: int = 1024

GT4PY_COMPILE_OPT_LEVEL = os.environ.get("GT4PY_COMPILE_OPT_LEVEL", "3")
_cxx_compiler_infos = cxx_compiler_defaults(GT4PY_COMPILE_OPT_LEVEL)
_gpu_compiler_configuration = gpu_configuration(GT4PY_COMPILE_OPT_LEVEL)

GT4PY_EXTRA_COMPILE_OPT_FLAGS = os.environ.get("GT4PY_EXTRA_COMPILE_OPT_FLAGS", "")

# Settings dict
GT4PY_EXTRA_COMPILE_ARGS = os.environ.get(
    "GT4PY_EXTRA_COMPILE_ARGS", _cxx_compiler_infos.cxx_compile_flags
)
_extra_compile_args = GT4PY_EXTRA_COMPILE_ARGS.split(" ") if GT4PY_EXTRA_COMPILE_ARGS else []

GT4PY_EXTRA_LINK_ARGS = os.environ.get("GT4PY_EXTRA_LINK_ARGS", "")
extra_link_args = GT4PY_EXTRA_LINK_ARGS.split(" ") if GT4PY_EXTRA_LINK_ARGS else []

# Resolve OpenMP
_enable_open_mp = os.environ.get(
    "GT4PY_CARTESIAN_ENABLE_OPENMP",
    "True" if _cxx_compiler_infos.enable_openmp else "False",
)
GT4PY_CARTESIAN_ENABLE_OPENMP = _enable_open_mp.lower() not in [
    "0",
    "false",
    "off",
]
if GT4PY_CARTESIAN_ENABLE_OPENMP:
    _openmp_cppflags = os.environ.get("OPENMP_CPPFLAGS", _cxx_compiler_infos.open_mp_flag).split()
    _openmp_ldflags = os.environ.get("OPENMP_LDFLAGS", _cxx_compiler_infos.open_mp_flag).split()
else:
    _openmp_cppflags = []
    _openmp_ldflags = []

build_settings: dict[str, Any] = {
    "cuda_bin_path": _gpu_compiler_configuration.binary_path,
    "cuda_include_path": _gpu_compiler_configuration.include_path,
    "cuda_library_path": _gpu_compiler_configuration.library_path,
    "cuda_arch": _gpu_compiler_configuration.arch,
    "gt_include_path": os.environ.get("GT_INCLUDE_PATH", GT_INCLUDE_PATH),
    "openmp": {
        "use_openmp": GT4PY_CARTESIAN_ENABLE_OPENMP,
        "cppflags": _openmp_cppflags,
        "ldflags": _openmp_ldflags,
    },
    "extra_compile_args": {
        "cxx": _extra_compile_args,
        "cuda": _gpu_compiler_configuration.gpu_compile_flags,
    },
    "extra_link_args": extra_link_args,
    "parallel_jobs": multiprocessing.cpu_count(),
    "cpp_template_depth": os.environ.get("GT_CPP_TEMPLATE_DEPTH", GT_CPP_TEMPLATE_DEPTH),
}

if _gpu_compiler_configuration.host_compiler is not None:
    build_settings["extra_compile_args"]["cuda"].append(
        f"-ccbin={_gpu_compiler_configuration.host_compiler}"
    )

cache_settings: dict[str, Any] = {
    "dir_name": os.environ.get("GT_CACHE_DIR_NAME", ".gt_cache"),
    "root_path": os.environ.get("GT_CACHE_ROOT", os.path.abspath(".")),
    "load_retries": int(os.environ.get("GT_CACHE_LOAD_RETRIES", 3)),
    "load_retry_delay": int(os.environ.get("GT_CACHE_LOAD_RETRY_DELAY", 100)),  # unit milliseconds
}

code_settings: dict[str, Any] = {"root_package_name": "_GT_"}

os.environ.setdefault("DACE_CONFIG", os.path.join(os.path.abspath("."), ".dace.conf"))

DACE_DEFAULT_BLOCK_SIZE = os.environ.get("DACE_DEFAULT_BLOCK_SIZE", "64,8,1")
